import rich
import rich_click as click

from union.cli._app_common import DeployApplicationGroupForFiles
from union.cli._common import CommandBase
from union.remote import UnionRemote


@click.group(name="deploy")
def deploy():
    """Deploy a resource."""


app_help = """Deploy application on Union."""
app_group = DeployApplicationGroupForFiles(
    name="apps",
    help=app_help,
    command_name="deploy",
)
deploy.add_command(app_group)


@deploy.command(cls=CommandBase)
@click.argument("model_uri")
@click.option(
    "--engine",
    "-e",
    type=str,
    default="auto",
    help="Specify the engine to use for deployment. If not provided, an engine will be automatically selected."
    "Ensure the value corresponds to one of the supported engines; otherwise, an error will be returned.",
)
@click.option("--dry-run", is_flag=True, help="Print the app spec without deploying.")
def model(model_uri: str, project: str, domain: str, engine: str, dry_run: bool):
    """Deploy model on Union."""
    remote = UnionRemote(default_project=project, default_domain=domain)
    deployable_model = remote._get_deployable_model(model_uri)
    console = rich.console.Console()

    engines = []
    preferred_engine = None
    for t in deployable_model.app_templates:
        engines.append(t.engine)
        if t.preferred:
            preferred_engine = t.engine

    console.print(f"Engines: {engines}, Preferred Engine: {preferred_engine}")
    if engine != "auto":
        if engine not in engines:
            raise click.BadParameter(
                f"Engine {engine} is not supported for this model. Supported engines are: {engines}"
            )
    else:
        engine = preferred_engine

    console.print(f"Deploying model {model_uri} using engine {engine}")
    app_template = next(t for t in deployable_model.app_templates if t.engine == engine)
    app = app_template.app
    app.container_image = remote._get_deployable_engine_container(engine)

    if dry_run:
        console.print(app)
        return
    app = remote.deploy_app(app)



import logging
import re
import typing
from pathlib import Path
from typing import Optional

import rich.markup
import rich_click as click
import yaml
from flytekit.configuration import (
    AuthType,
    Config,
    ImageConfig,
    PlatformConfig,
    SerializationSettings,
)
from flytekit.remote import FlyteTask, remote_fs
from flytekit.tools.translator import get_serializable
from google.protobuf.json_format import MessageToDict
from grpc import RpcError
from rich.console import Console

from union import Artifact
from union._config import (
    AppClientCredentials,
    _encode_app_client_credentials,
    _get_config_obj,
    _get_endpoint_for_login,
    _get_user_handle,
    _is_serverless_endpoint,
    _write_config_to_path,
)
from union.cli._artifact_create_common import ArtifactCreateCommand
from union.cli._common import _get_channel_with_org, _validate_key_value_pairs
from union.cli._option import MutuallyExclusiveArgument, MutuallyExclusiveOption
from union.internal.identity.app_payload_pb2 import CreateAppRequest
from union.internal.identity.app_service_pb2_grpc import AppsServiceStub
from union.internal.identity.enums_pb2 import (
    ConsentMethod,
    GrantTypes,
    ResponseTypes,
    TokenEndpointAuthMethod,
)
from union.internal.secret.definition_pb2 import SecretIdentifier, SecretSpec
from union.internal.secret.payload_pb2 import CreateSecretRequest
from union.internal.secret.secret_pb2_grpc import SecretServiceStub
from union.remote import UnionRemote
from union.workspace._config import _DEFAULT_CONFIG_YAML_FOR_BASE_IMAGE, _DEFAULT_CONFIG_YAML_FOR_IMAGE_SPEC
from union.workspace._vscode import WorkspaceConfig, resolver
from union.workspace._vscode_remote import _generate_workspace_name

logger = logging.getLogger("union.cli._create")


@click.group()
def create():
    """Create a resource."""


@create.command()
@click.argument(
    "name",
    required=False,
    cls=MutuallyExclusiveArgument,
    mutually_exclusive=["name_option"],
    error_msg="Please pass --name once: `union create secret --name NAME`",
)
@click.option(
    "--name",
    "name_option",
    help="Secret name",
    cls=MutuallyExclusiveOption,
    mutually_exclusive=["name"],
    error_msg="Please pass --name once: `union create secret --name NAME`",
)
@click.option(
    "--value",
    help="Secret value",
    prompt="Enter secret value",
    hide_input=True,
    cls=MutuallyExclusiveOption,
    mutually_exclusive=["value_file"],
)
@click.option(
    "-f",
    "--value-file",
    help="Path to file containing the secret",
    type=click.Path(exists=True, file_okay=True, dir_okay=False, readable=True, allow_dash=True),
    cls=MutuallyExclusiveOption,
    mutually_exclusive=["value"],
)
@click.option("--project", help="Project name")
@click.option("--domain", help="Domain name")
def secret(
    name: Optional[str],
    name_option: Optional[str],
    value: str,
    value_file: str,
    project: Optional[str],
    domain: Optional[str],
):
    """Create a secret with NAME."""
    name = name or name_option

    platform_obj = _get_config_obj().platform
    channel, org = _get_channel_with_org(platform_obj)

    if value_file:
        with open(value_file, "rb") as f:
            secret_spec = SecretSpec(binary_value=f.read())
    else:
        secret_spec = SecretSpec(string_value=value)

    stub = SecretServiceStub(channel)
    request = CreateSecretRequest(
        id=SecretIdentifier(name=name, domain=domain, project=project, organization=org),
        secret_spec=secret_spec,
    )
    try:
        stub.CreateSecret(request)
        click.echo(f"Created secret with name: {name}")
    except RpcError as e:
        raise click.ClickException(f"Unable to create secret with name: {name}\n{e}") from e


@create.group("api-key")
def api_key():
    """Manage API keys."""


@api_key.command("admin")
@click.option("--name", type=str, help="Name for API key", required=True)
def admin(name: str):
    """Create an api key."""
    platform_obj = _get_config_obj().platform
    normalized_client_name = re.sub("[^0-9a-zA-Z]+", "-", name.lower())
    if _is_serverless_endpoint(platform_obj.endpoint):
        userhandle = _get_user_handle(platform_obj)
        tenant = platform_obj.endpoint.split(".")[0]
        client_id = f"{tenant}-{userhandle}-{normalized_client_name}"
    else:
        client_id = normalized_client_name
    channel, org = _get_channel_with_org(platform_obj)

    stub = AppsServiceStub(channel)
    request = CreateAppRequest(
        organization=org,
        client_id=client_id,
        client_name=client_id,
        grant_types=[GrantTypes.CLIENT_CREDENTIALS, GrantTypes.AUTHORIZATION_CODE],
        redirect_uris=["http://localhost:8080/authorization-code/callback"],
        response_types=[ResponseTypes.CODE],
        token_endpoint_auth_method=TokenEndpointAuthMethod.CLIENT_SECRET_BASIC,
        consent_method=ConsentMethod.CONSENT_METHOD_REQUIRED,
    )

    try:
        response = stub.Create(request)
    except RpcError as e:
        raise click.ClickException(f"Unable to create api-key with name: {name}\n{e}") from e

    click.echo(f"Client ID: {response.app.client_id}")
    click.echo("The following API key will only be shown once. Be sure to keep it safe!")
    click.echo("Configure your headless CLI by setting the following environment variable:")
    click.echo()

    union_api_key = _encode_app_client_credentials(
        AppClientCredentials(
            endpoint=platform_obj.endpoint,
            client_id=response.app.client_id,
            client_secret=response.app.client_secret,
            org=org,
        )
    )
    click.echo(f'export UNION_API_KEY="{union_api_key}"')


@create.command()
@click.option(
    "--auth",
    type=click.Choice(["device-flow", "pkce"]),
    default="pkce",
    help="Authorization method to ues",
)
@click.option(
    "--host", default=None, help="Host to connect to.", cls=MutuallyExclusiveOption, mutually_exclusive=["serverless"]
)
@click.option(
    "--serverless",
    default=False,
    is_flag=True,
    help="Connect to serverless.",
    cls=MutuallyExclusiveOption,
    mutually_exclusive=["host"],
)
def login(auth: str, host: Optional[str], serverless: bool):
    """Log into Union

    - On Union Serverless run: `union create login --serverless`

    - On Union BYOC run: `union create login --host UNION_TENANT`
    """
    endpoint = _get_endpoint_for_login(host=host, serverless=serverless)

    if auth == "pkce":
        auth_mode = AuthType.PKCE
    else:
        auth_mode = AuthType.DEVICEFLOW

    config = Config.auto().with_params(
        platform=PlatformConfig(
            endpoint=endpoint,
            insecure=False,
            auth_mode=auth_mode,
        )
    )

    console = Console()
    try:
        path = _write_config_to_path(endpoint, auth_mode.value)
        console.print(f"🔐 [yellow]Configuration saved to {path}[/yellow]")

        # Accessing the client will trigger authentication
        remote = UnionRemote(config=config)
        remote._user_info()  # noqa
        ep = "serverless" if serverless else endpoint
        console.print(f"Login successful into [green][bold]{ep}[/bold][/green]")

    except Exception as e:
        raise click.ClickException(f"Unable to login.\n{e}") from e


@create.command()
@click.pass_context
@click.argument("config_file", type=click.Path(path_type=Path))
def workspace(ctx: click.Context, config_file: Path):
    """Create workspace."""
    config_content = config_file.read_text()
    workspace_config = WorkspaceConfig.from_yaml(config_content)

    remote = UnionRemote(default_domain=workspace_config.domain, default_project=workspace_config.project)
    logger.debug(f"Workspace config: {workspace_config}")
    logger.debug(f"Remote: {remote.__dict__}")
    if not isinstance(workspace_config.container_image, str):
        _image_build_task = _build_workspace_image(remote, workspace_config)
        # override container image with the built image
        workspace_config.container_image = _image_build_task.template.container.image
    _create_workspace(remote, workspace_config)


@create.command("workspace-config")
@click.argument("config_file", type=click.Path(path_type=Path))
@click.option("--init", type=click.Choice(["base_image", "custom_image"]), required=True)
def workspace_config(config_file: Path, init: str):
    """Create workspace config at CONFIG_FILE."""
    from union._config import _is_serverless_endpoint

    if config_file.exists():
        raise click.ClickException(f"{config_file} already exists")

    remote = UnionRemote()
    if _is_serverless_endpoint(remote.config.platform.endpoint):
        default_project = "default"
    else:
        default_project = "flytesnacks"

    click.echo(f"Writing workspace configuration to {config_file}")
    if init == "base_image":
        config_file.write_text(_DEFAULT_CONFIG_YAML_FOR_BASE_IMAGE.format(default_project=default_project))
    else:
        config_file.write_text(_DEFAULT_CONFIG_YAML_FOR_IMAGE_SPEC.format(default_project=default_project))


def _build_workspace_image(
    remote: UnionRemote, ws_config: WorkspaceConfig, dry_run: bool = False
) -> typing.Optional[FlyteTask]:
    name = ws_config.name
    project = ws_config.project
    domain = ws_config.domain

    task = resolver.get_task_for_workspace(ws_config)
    console = Console()
    console.print("✨ Building workspace image...")

    try:
        source_root = task.container_image.source_root
    except AttributeError:
        source_root = None

    serialization_settings = SerializationSettings(
        project=project, domain=domain, image_config=ImageConfig.auto_default_image(), source_root=source_root
    )
    if dry_run:
        m = typing.OrderedDict()
        task_entity = get_serializable(m, settings=serialization_settings, entity=task)
        yaml_output = yaml.safe_dump(
            MessageToDict(task_entity.to_flyte_idl()), sort_keys=False, default_flow_style=False
        )
        console.print("Would register:")
        console.print(yaml_output)
        return

    return remote.register_task(
        task,
        serialization_settings=serialization_settings,
        version=_generate_workspace_name(name),
    )


def _create_workspace(remote: UnionRemote, ws_config: WorkspaceConfig):
    from union._async import run_sync
    from union.internal.workspace.workspace_definition_payload_pb2 import CreateWorkspaceDefinitionResponse
    from union.workspace._vscode_remote import WorkspaceRemote

    name = ws_config.name
    ws_remote = WorkspaceRemote(
        default_project=ws_config.project,
        default_domain=ws_config.domain,
        union_remote=remote,
    )

    console = Console()
    ws_def: CreateWorkspaceDefinitionResponse = run_sync(ws_remote.create_workspace_definition, ws_config)

    console.print(f"Created: {ws_def}")
    console.print(f"Starting workspace '{name}'")
    with console.status("Provisioning workspace", spinner="dots"):
        try:
            response = run_sync(ws_remote.start_workspace_instance, ws_def.workspace_definition.id.name)
        except Exception as e:
            console.print(f"Error provisioning workspace: {rich.markup.escape(e.details())}", style="red")
            raise e

    url = f"https://{remote.config.platform.endpoint}/{response.workspace_instance.spec.uri}"
    loading_url = f"https://{remote.config.platform.endpoint}/loading?type=workspace&url={url}"
    console.print(f"\n🚀 Workspace started: [link={loading_url}]Open VSCode in Browser[/link]")


@create.command("artifact", cls=ArtifactCreateCommand)
@click.argument("name")
@click.option("--version", type=str, required=True, help="Version of the artifact")
@click.option(
    "--partitions", "-p", callback=_validate_key_value_pairs, help="Partitions for the artifact", multiple=True
)
@click.option("--short-description", help="Short description of the artifact")
def artifact(
    name: str,
    project: str,
    domain: str,
    version: Optional[str],
    partitions: Optional[typing.List[str]],
    short_description: Optional[str],
    **kwargs,
):
    """Create an artifact with NAME."""
    remote = UnionRemote(
        default_domain=domain, default_project=project, data_upload_location=remote_fs.REMOTE_PLACEHOLDER
    )
    lit, lt = ArtifactCreateCommand.get_literal_from_args(remote.context, kwargs)
    a = Artifact(
        project=project,
        domain=domain,
        name=name,
        version=version,
        partitions=partitions,
        short_description=short_description,
        literal=lit,
        literal_type=lt,
    )
    remote.create_artifact(artifact=a)
    url = remote.generate_console_url(a)
    Console().print(f"[green]Created artifact with name: [bold][link={url}]{name}:{version}[/link][/bold][/green]")



from union.internal.common import list_pb2 as _list_pb2
from flyteidl.admin import task_pb2 as _task_pb2
from union.internal.validate.validate import validate_pb2 as _validate_pb2
from union.internal.workspace import workspace_definition_pb2 as _workspace_definition_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class CreateWorkspaceDefinitionRequest(_message.Message):
    __slots__ = ["id", "spec", "short_description", "workspace_spec"]
    ID_FIELD_NUMBER: _ClassVar[int]
    SPEC_FIELD_NUMBER: _ClassVar[int]
    SHORT_DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    WORKSPACE_SPEC_FIELD_NUMBER: _ClassVar[int]
    id: _workspace_definition_pb2.WorkspaceDefinitionIdentifier
    spec: _task_pb2.TaskSpec
    short_description: str
    workspace_spec: _workspace_definition_pb2.WorkspaceDefinitionSpec
    def __init__(self, id: _Optional[_Union[_workspace_definition_pb2.WorkspaceDefinitionIdentifier, _Mapping]] = ..., spec: _Optional[_Union[_task_pb2.TaskSpec, _Mapping]] = ..., short_description: _Optional[str] = ..., workspace_spec: _Optional[_Union[_workspace_definition_pb2.WorkspaceDefinitionSpec, _Mapping]] = ...) -> None: ...

class CreateWorkspaceDefinitionResponse(_message.Message):
    __slots__ = ["workspace_definition"]
    WORKSPACE_DEFINITION_FIELD_NUMBER: _ClassVar[int]
    workspace_definition: _workspace_definition_pb2.WorkspaceDefinition
    def __init__(self, workspace_definition: _Optional[_Union[_workspace_definition_pb2.WorkspaceDefinition, _Mapping]] = ...) -> None: ...

class GetWorkspaceDefinitionRequest(_message.Message):
    __slots__ = ["id"]
    ID_FIELD_NUMBER: _ClassVar[int]
    id: _workspace_definition_pb2.WorkspaceDefinitionIdentifier
    def __init__(self, id: _Optional[_Union[_workspace_definition_pb2.WorkspaceDefinitionIdentifier, _Mapping]] = ...) -> None: ...

class GetWorkspaceDefinitionResponse(_message.Message):
    __slots__ = ["workspace_definition"]
    WORKSPACE_DEFINITION_FIELD_NUMBER: _ClassVar[int]
    workspace_definition: _workspace_definition_pb2.WorkspaceDefinition
    def __init__(self, workspace_definition: _Optional[_Union[_workspace_definition_pb2.WorkspaceDefinition, _Mapping]] = ...) -> None: ...

class UpdateWorkspaceDefinitionRequest(_message.Message):
    __slots__ = ["id", "spec", "state", "short_description", "workspace_spec"]
    ID_FIELD_NUMBER: _ClassVar[int]
    SPEC_FIELD_NUMBER: _ClassVar[int]
    STATE_FIELD_NUMBER: _ClassVar[int]
    SHORT_DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    WORKSPACE_SPEC_FIELD_NUMBER: _ClassVar[int]
    id: _workspace_definition_pb2.WorkspaceDefinitionIdentifier
    spec: _task_pb2.TaskSpec
    state: _workspace_definition_pb2.WorkspaceDefinitionState
    short_description: str
    workspace_spec: _workspace_definition_pb2.WorkspaceDefinitionSpec
    def __init__(self, id: _Optional[_Union[_workspace_definition_pb2.WorkspaceDefinitionIdentifier, _Mapping]] = ..., spec: _Optional[_Union[_task_pb2.TaskSpec, _Mapping]] = ..., state: _Optional[_Union[_workspace_definition_pb2.WorkspaceDefinitionState, str]] = ..., short_description: _Optional[str] = ..., workspace_spec: _Optional[_Union[_workspace_definition_pb2.WorkspaceDefinitionSpec, _Mapping]] = ...) -> None: ...

class UpdateWorkspaceDefinitionResponse(_message.Message):
    __slots__ = ["workspace_definition"]
    WORKSPACE_DEFINITION_FIELD_NUMBER: _ClassVar[int]
    workspace_definition: _workspace_definition_pb2.WorkspaceDefinition
    def __init__(self, workspace_definition: _Optional[_Union[_workspace_definition_pb2.WorkspaceDefinition, _Mapping]] = ...) -> None: ...

class ListWorkspaceDefinitionsRequest(_message.Message):
    __slots__ = ["org", "project", "domain", "name", "request"]
    ORG_FIELD_NUMBER: _ClassVar[int]
    PROJECT_FIELD_NUMBER: _ClassVar[int]
    DOMAIN_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    REQUEST_FIELD_NUMBER: _ClassVar[int]
    org: str
    project: str
    domain: str
    name: str
    request: _list_pb2.ListRequest
    def __init__(self, org: _Optional[str] = ..., project: _Optional[str] = ..., domain: _Optional[str] = ..., name: _Optional[str] = ..., request: _Optional[_Union[_list_pb2.ListRequest, _Mapping]] = ...) -> None: ...

class ListWorkspaceDefinitionsResponse(_message.Message):
    __slots__ = ["workspace_definitions", "token"]
    WORKSPACE_DEFINITIONS_FIELD_NUMBER: _ClassVar[int]
    TOKEN_FIELD_NUMBER: _ClassVar[int]
    workspace_definitions: _containers.RepeatedCompositeFieldContainer[_workspace_definition_pb2.WorkspaceDefinition]
    token: str
    def __init__(self, workspace_definitions: _Optional[_Iterable[_Union[_workspace_definition_pb2.WorkspaceDefinition, _Mapping]]] = ..., token: _Optional[str] = ...) -> None: ...

class ListWorkspaceDefinitionEntitiesRequest(_message.Message):
    __slots__ = ["org", "project", "domain", "request"]
    ORG_FIELD_NUMBER: _ClassVar[int]
    PROJECT_FIELD_NUMBER: _ClassVar[int]
    DOMAIN_FIELD_NUMBER: _ClassVar[int]
    REQUEST_FIELD_NUMBER: _ClassVar[int]
    org: str
    project: str
    domain: str
    request: _list_pb2.ListRequest
    def __init__(self, org: _Optional[str] = ..., project: _Optional[str] = ..., domain: _Optional[str] = ..., request: _Optional[_Union[_list_pb2.ListRequest, _Mapping]] = ...) -> None: ...

class ListWorkspaceDefinitionEntitiesResponse(_message.Message):
    __slots__ = ["workspace_definition_entities", "token"]
    WORKSPACE_DEFINITION_ENTITIES_FIELD_NUMBER: _ClassVar[int]
    TOKEN_FIELD_NUMBER: _ClassVar[int]
    workspace_definition_entities: _containers.RepeatedCompositeFieldContainer[_workspace_definition_pb2.WorkspaceDefinitionEntity]
    token: str
    def __init__(self, workspace_definition_entities: _Optional[_Iterable[_Union[_workspace_definition_pb2.WorkspaceDefinitionEntity, _Mapping]]] = ..., token: _Optional[str] = ...) -> None: ...

class UpdateWorkspaceDefinitionEntityRequest(_message.Message):
    __slots__ = ["entity"]
    ENTITY_FIELD_NUMBER: _ClassVar[int]
    entity: _workspace_definition_pb2.WorkspaceDefinitionEntity
    def __init__(self, entity: _Optional[_Union[_workspace_definition_pb2.WorkspaceDefinitionEntity, _Mapping]] = ...) -> None: ...

class UpdateWorkspaceDefinitionEntityResponse(_message.Message):
    __slots__ = []
    def __init__(self) -> None: ...



import math
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Set, Tuple, Union

from flyteidl.core import workflow_pb2 as _core_workflow
from flytekit.core import interface as flyte_interface
from flytekit.core.array_node import ArrayNode
from flytekit.core.context_manager import ExecutionState, FlyteContext
from flytekit.core.interface import (
    transform_interface_to_list_interface,
    transform_interface_to_typed_interface,
)
from flytekit.core.launch_plan import LaunchPlan
from flytekit.core.node import Node
from flytekit.core.promise import (
    Promise,
    VoidPromise,
    create_and_link_node,
    create_and_link_node_from_remote,
    flyte_entity_call_handler,
    translate_inputs_to_literals,
)
from flytekit.core.task import ReferenceTask
from flytekit.loggers import logger
from flytekit.models import interface as _interface_models
from flytekit.models import literals as _literal_models
from flytekit.models.core import workflow as _workflow_model
from flytekit.models.literals import Literal, LiteralCollection, Scalar

ARRAY_NODE_SUBNODE_NAME = "array_node_subnode"


if TYPE_CHECKING:
    from flytekit.remote import FlyteLaunchPlan


class UnionArrayNode(ArrayNode):
    def __init__(
        self,
        target: Union[LaunchPlan, ReferenceTask, "FlyteLaunchPlan"],
        bound_inputs: Optional[Dict[str, Any]] = None,
        bindings: Optional[List[_literal_models.Binding]] = None,
        concurrency: Optional[int] = None,
        min_successes: Optional[int] = None,
        min_success_ratio: Optional[float] = None,
        metadata: Optional[_workflow_model.NodeMetadata] = None,
    ):
        """
        :param target: The target Flyte entity to map over
        :param bound_inputs: Inputs that are bound to the array node and will not be mapped over
        :param concurrency: If specified, this limits the number of mapped tasks than can run in parallel to the given
            batch size. If the size of the input exceeds the concurrency value, then multiple batches will be run
            serially until all inputs are processed. If set to 0, this means unbounded concurrency. If left unspecified,
            this means the array node will inherit parallelism from the workflow
        :param min_successes: The minimum number of successful executions. If set, this takes precedence over
            min_success_ratio
        :param min_success_ratio: The minimum ratio of successful executions.
        :param metadata: The metadata for the underlying node
        """
        from flytekit.remote import FlyteLaunchPlan

        self.target = target
        self._concurrency = concurrency
        self.id = target.name
        self._bindings = bindings or []
        self._bound_inputs = bound_inputs or {}
        self.metadata = metadata
        self._data_mode = None
        self._execution_mode = None

        if min_successes is not None:
            self._min_successes = min_successes
            self._min_success_ratio = None
        else:
            self._min_success_ratio = min_success_ratio if min_success_ratio is not None else 1.0
            self._min_successes = 0

        if self.target.python_interface:
            n_outputs = len(self.target.python_interface.outputs)
        else:
            n_outputs = len(self.target.interface.outputs)
        if n_outputs > 1:
            raise ValueError("Only tasks with a single output are supported in map tasks.")

        output_as_list_of_optionals = min_success_ratio is not None and min_success_ratio != 1 and n_outputs == 1

        self._remote_interface = None
        if self.target.python_interface:
            self._python_interface = transform_interface_to_list_interface(
                self.target.python_interface, self.bound_inputs, set(), output_as_list_of_optionals
            )
        elif self.target.interface:
            self._remote_interface = self.target.interface.transform_interface_to_list(self.bound_inputs, set())
        else:
            raise ValueError("No interface found for the target entity.")

        if isinstance(target, (LaunchPlan, FlyteLaunchPlan)):
            self._data_mode = _core_workflow.ArrayNode.SINGLE_INPUT_FILE
            self._execution_mode = _core_workflow.ArrayNode.FULL_STATE
        elif isinstance(target, ReferenceTask):
            self._data_mode = _core_workflow.ArrayNode.INDIVIDUAL_INPUT_FILES
            self._execution_mode = _core_workflow.ArrayNode.MINIMAL_STATE
        else:
            raise ValueError(f"Only LaunchPlans are supported for now, but got {type(target)}")

    def construct_node_metadata(self) -> _workflow_model.NodeMetadata:
        # Part of SupportsNodeCreation interface
        return self.metadata or _workflow_model.NodeMetadata(name=self.target.name)

    @property
    def name(self) -> str:
        # Part of SupportsNodeCreation interface
        return self.target.name

    @property
    def python_interface(self) -> flyte_interface.Interface:
        # Part of SupportsNodeCreation interface
        return self._python_interface

    @property
    def interface(self) -> _interface_models.TypedInterface:
        # Required in get_serializable_node
        if self._remote_interface:
            return self._remote_interface
        raise AttributeError("interface attribute is not available")

    @property
    def bindings(self) -> List[_literal_models.Binding]:
        # Required in get_serializable_node
        return self._bindings

    @property
    def upstream_nodes(self) -> List[Node]:
        # Required in get_serializable_node
        return []

    @property
    def flyte_entity(self) -> Any:
        return self.target

    @property
    def data_mode(self) -> _core_workflow.ArrayNode.DataMode:
        return self._data_mode

    def local_execute(self, ctx: FlyteContext, **kwargs) -> Union[Tuple[Promise], Promise, VoidPromise]:
        if self._remote_interface:
            raise ValueError("Mapping over remote entities is not supported in local execution.")

        outputs_expected = True
        if not self.python_interface.outputs:
            outputs_expected = False

        mapped_entity_count = 0
        for binding in self.bindings:
            k = binding.var
            if k not in self._bound_inputs:
                v = kwargs[k]
                if isinstance(v, list) and len(v) > 0 and isinstance(v[0], self.target.python_interface.inputs[k]):
                    mapped_entity_count = len(v)
                    break
                else:
                    raise ValueError(
                        f"Expected a list of {self.target.python_interface.inputs[k]} but got {type(v)} instead."
                    )

        failed_count = 0
        min_successes = mapped_entity_count
        if self._min_successes:
            min_successes = self._min_successes
        elif self._min_success_ratio:
            min_successes = math.ceil(min_successes * self._min_success_ratio)

        literals = []
        for i in range(mapped_entity_count):
            single_instance_inputs = {}
            for binding in self.bindings:
                k = binding.var
                if k not in self._bound_inputs:
                    single_instance_inputs[k] = kwargs[k][i]
                else:
                    single_instance_inputs[k] = kwargs[k]

            # translate Python native inputs to Flyte literals
            typed_interface = transform_interface_to_typed_interface(self.target.python_interface)
            literal_map = translate_inputs_to_literals(
                ctx,
                incoming_values=single_instance_inputs,
                flyte_interface_types={} if typed_interface is None else typed_interface.inputs,
                native_types=self.target.python_interface.inputs,
            )
            kwargs_literals = {k1: Promise(var=k1, val=v1) for k1, v1 in literal_map.items()}

            try:
                output = self.target.__call__(**kwargs_literals)
                if outputs_expected:
                    literals.append(output.val)
            except Exception as exc:
                if outputs_expected:
                    literal_with_none = Literal(scalar=Scalar(none_type=_literal_models.Void()))
                    literals.append(literal_with_none)
                    failed_count += 1
                    if mapped_entity_count - failed_count < min_successes:
                        logger.error("The number of successful tasks is lower than the minimum")
                        raise exc

        if outputs_expected:
            return Promise(var="o0", val=Literal(collection=LiteralCollection(literals=literals)))
        return VoidPromise(self.name)

    def local_execution_mode(self):
        return ExecutionState.Mode.LOCAL_TASK_EXECUTION

    @property
    def min_success_ratio(self) -> Optional[float]:
        return self._min_success_ratio

    @property
    def min_successes(self) -> Optional[int]:
        return self._min_successes

    @property
    def concurrency(self) -> Optional[int]:
        return self._concurrency

    @property
    def execution_mode(self) -> _core_workflow.ArrayNode.ExecutionMode:
        return self._execution_mode

    @property
    def is_original_sub_node_interface(self) -> bool:
        return True

    @property
    def bound_inputs(self) -> Set[str]:
        return set(self._bound_inputs)

    def __call__(self, *args, **kwargs):
        kwargs.update(self._bound_inputs)
        if not self._bindings:
            ctx = FlyteContext.current_context()
            # since a new entity with an updated list interface is not created, we have to work around the mismatch
            # between the interface and the inputs. Also, don't link the node to the compilation state,
            # since we don't want to add the subnode to the workflow as a node
            if self._remote_interface:
                bound_subnode = create_and_link_node_from_remote(
                    ctx,
                    entity=self.flyte_entity,
                    add_node_to_compilation_state=False,
                    overridden_interface=self._remote_interface,
                    **kwargs,
                )
                self._bindings = bound_subnode.ref.node.bindings
                return create_and_link_node_from_remote(
                    ctx,
                    entity=self,
                    **kwargs,
                )
            bound_subnode = create_and_link_node(
                ctx,
                entity=self.flyte_entity,
                add_node_to_compilation_state=False,
                overridden_interface=self.python_interface,
                node_id=ARRAY_NODE_SUBNODE_NAME,
                **kwargs,
            )
            self._bindings = bound_subnode.ref.node.bindings
        return flyte_entity_call_handler(self, *args, **kwargs)



from union.internal.common import identity_pb2 as _identity_pb2
from flyteidl.admin import task_pb2 as _task_pb2
from flyteidl.core import identifier_pb2 as _identifier_pb2
from flyteidl.core import security_pb2 as _security_pb2
from flyteidl.core import tasks_pb2 as _tasks_pb2
from google.protobuf import timestamp_pb2 as _timestamp_pb2
from union.internal.validate.validate import validate_pb2 as _validate_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class WorkspaceDefinitionState(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []
    WORKSPACE_DEFINITION_STATE_UNSPECIFIED: _ClassVar[WorkspaceDefinitionState]
    WORKSPACE_DEFINITION_STATE_ACTIVE: _ClassVar[WorkspaceDefinitionState]
    WORKSPACE_DEFINITION_STATE_ARCHIVED: _ClassVar[WorkspaceDefinitionState]
WORKSPACE_DEFINITION_STATE_UNSPECIFIED: WorkspaceDefinitionState
WORKSPACE_DEFINITION_STATE_ACTIVE: WorkspaceDefinitionState
WORKSPACE_DEFINITION_STATE_ARCHIVED: WorkspaceDefinitionState

class WorkspaceDefinitionIdentifier(_message.Message):
    __slots__ = ["org", "project", "domain", "name", "version"]
    ORG_FIELD_NUMBER: _ClassVar[int]
    PROJECT_FIELD_NUMBER: _ClassVar[int]
    DOMAIN_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    VERSION_FIELD_NUMBER: _ClassVar[int]
    org: str
    project: str
    domain: str
    name: str
    version: str
    def __init__(self, org: _Optional[str] = ..., project: _Optional[str] = ..., domain: _Optional[str] = ..., name: _Optional[str] = ..., version: _Optional[str] = ...) -> None: ...

class WorkspaceDefinition(_message.Message):
    __slots__ = ["id", "spec"]
    ID_FIELD_NUMBER: _ClassVar[int]
    SPEC_FIELD_NUMBER: _ClassVar[int]
    id: WorkspaceDefinitionIdentifier
    spec: WorkspaceDefinitionSpec
    def __init__(self, id: _Optional[_Union[WorkspaceDefinitionIdentifier, _Mapping]] = ..., spec: _Optional[_Union[WorkspaceDefinitionSpec, _Mapping]] = ...) -> None: ...

class WorkspaceDefinitionSpec(_message.Message):
    __slots__ = ["short_description", "created_at", "updated_at", "task_id", "state", "task_spec", "resources", "extended_resources", "ttl_seconds", "security_context", "python_environment", "conda_environment", "on_startup", "container_image"]
    SHORT_DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    CREATED_AT_FIELD_NUMBER: _ClassVar[int]
    UPDATED_AT_FIELD_NUMBER: _ClassVar[int]
    TASK_ID_FIELD_NUMBER: _ClassVar[int]
    STATE_FIELD_NUMBER: _ClassVar[int]
    TASK_SPEC_FIELD_NUMBER: _ClassVar[int]
    RESOURCES_FIELD_NUMBER: _ClassVar[int]
    EXTENDED_RESOURCES_FIELD_NUMBER: _ClassVar[int]
    TTL_SECONDS_FIELD_NUMBER: _ClassVar[int]
    SECURITY_CONTEXT_FIELD_NUMBER: _ClassVar[int]
    PYTHON_ENVIRONMENT_FIELD_NUMBER: _ClassVar[int]
    CONDA_ENVIRONMENT_FIELD_NUMBER: _ClassVar[int]
    ON_STARTUP_FIELD_NUMBER: _ClassVar[int]
    CONTAINER_IMAGE_FIELD_NUMBER: _ClassVar[int]
    short_description: str
    created_at: _timestamp_pb2.Timestamp
    updated_at: _timestamp_pb2.Timestamp
    task_id: _identifier_pb2.Identifier
    state: WorkspaceDefinitionState
    task_spec: _task_pb2.TaskSpec
    resources: _tasks_pb2.Resources
    extended_resources: _tasks_pb2.ExtendedResources
    ttl_seconds: int
    security_context: _security_pb2.SecurityContext
    python_environment: PythonEnvironment
    conda_environment: CondaEnvironment
    on_startup: _containers.RepeatedScalarFieldContainer[str]
    container_image: str
    def __init__(self, short_description: _Optional[str] = ..., created_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., updated_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., task_id: _Optional[_Union[_identifier_pb2.Identifier, _Mapping]] = ..., state: _Optional[_Union[WorkspaceDefinitionState, str]] = ..., task_spec: _Optional[_Union[_task_pb2.TaskSpec, _Mapping]] = ..., resources: _Optional[_Union[_tasks_pb2.Resources, _Mapping]] = ..., extended_resources: _Optional[_Union[_tasks_pb2.ExtendedResources, _Mapping]] = ..., ttl_seconds: _Optional[int] = ..., security_context: _Optional[_Union[_security_pb2.SecurityContext, _Mapping]] = ..., python_environment: _Optional[_Union[PythonEnvironment, _Mapping]] = ..., conda_environment: _Optional[_Union[CondaEnvironment, _Mapping]] = ..., on_startup: _Optional[_Iterable[str]] = ..., container_image: _Optional[str] = ...) -> None: ...

class PythonEnvironment(_message.Message):
    __slots__ = ["packages"]
    PACKAGES_FIELD_NUMBER: _ClassVar[int]
    packages: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, packages: _Optional[_Iterable[str]] = ...) -> None: ...

class CondaEnvironment(_message.Message):
    __slots__ = ["packages", "channels"]
    PACKAGES_FIELD_NUMBER: _ClassVar[int]
    CHANNELS_FIELD_NUMBER: _ClassVar[int]
    packages: _containers.RepeatedScalarFieldContainer[str]
    channels: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, packages: _Optional[_Iterable[str]] = ..., channels: _Optional[_Iterable[str]] = ...) -> None: ...

class WorkspaceDefinitionEntityId(_message.Message):
    __slots__ = ["org", "project", "domain", "name"]
    ORG_FIELD_NUMBER: _ClassVar[int]
    PROJECT_FIELD_NUMBER: _ClassVar[int]
    DOMAIN_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    org: str
    project: str
    domain: str
    name: str
    def __init__(self, org: _Optional[str] = ..., project: _Optional[str] = ..., domain: _Optional[str] = ..., name: _Optional[str] = ...) -> None: ...

class AuditMetadata(_message.Message):
    __slots__ = ["created_at", "updated_at", "created_by", "updated_by"]
    CREATED_AT_FIELD_NUMBER: _ClassVar[int]
    UPDATED_AT_FIELD_NUMBER: _ClassVar[int]
    CREATED_BY_FIELD_NUMBER: _ClassVar[int]
    UPDATED_BY_FIELD_NUMBER: _ClassVar[int]
    created_at: _timestamp_pb2.Timestamp
    updated_at: _timestamp_pb2.Timestamp
    created_by: _identity_pb2.EnrichedIdentity
    updated_by: _identity_pb2.EnrichedIdentity
    def __init__(self, created_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., updated_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., created_by: _Optional[_Union[_identity_pb2.EnrichedIdentity, _Mapping]] = ..., updated_by: _Optional[_Union[_identity_pb2.EnrichedIdentity, _Mapping]] = ...) -> None: ...

class WorkspaceDefinitionEntity(_message.Message):
    __slots__ = ["id", "state", "short_description", "audit_metadata"]
    ID_FIELD_NUMBER: _ClassVar[int]
    STATE_FIELD_NUMBER: _ClassVar[int]
    SHORT_DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    AUDIT_METADATA_FIELD_NUMBER: _ClassVar[int]
    id: WorkspaceDefinitionEntityId
    state: WorkspaceDefinitionState
    short_description: str
    audit_metadata: AuditMetadata
    def __init__(self, id: _Optional[_Union[WorkspaceDefinitionEntityId, _Mapping]] = ..., state: _Optional[_Union[WorkspaceDefinitionState, str]] = ..., short_description: _Optional[str] = ..., audit_metadata: _Optional[_Union[AuditMetadata, _Mapping]] = ...) -> None: ...



"""This module contains utils that are useful for development.

Example usage:

```python
import pandas as pd
from flytekit import task, workflow

from union._testing import imagespec_with_local_union

image = imagespec_with_local_union(packages=["pandas"])


@task(container_image=image)
def get_data() -> pd.DataFrame:
    return pd.DataFrame([1, 2, 3])


@task(container_image=image)
def double_df(df: pd.DataFrame) -> pd.DataFrame:
    return 2 * df


@workflow
def main() -> pd.DataFrame:
    df = get_data()
    return double_df(df=df)
```
"""

import os
import sys
from pathlib import Path
from shutil import rmtree
from subprocess import run
from tempfile import mkdtemp
from typing import Optional

import flytekit
from flytekit import ImageSpec
from flytekit.constants import CopyFileDetection
from flytekit.core.context_manager import FlyteContextManager

import union


def imagespec_with_local_union(
    *args,
    builder: str,
    packages: Optional[list] = None,
    source_root: Optional[str] = None,
    pull_flytekit_local_dev=False,
    **kwargs,
) -> Optional[ImageSpec]:
    """Creates an imagespec with local changes."""
    state = FlyteContextManager.current_context().execution_state
    if not (state and state.mode is None):
        return None

    union_root_path = Path(union.__path__[0]).parent

    assert (union_root_path / "pyproject.toml").exists(), "pyproject.toml needs to exists in root directory"

    if source_root is None:
        tmp_path = Path(mkdtemp())
        source_root = tmp_path / "source_root"
        if source_root.exists():
            rmtree(source_root)
        source_root.mkdir(parents=True)
    else:
        source_root = Path(source_root)

    vendor_path = source_root / ".vendor"
    vendor_path = vendor_path.absolute()

    expected_wheels = 1
    run([sys.executable, "-m", "build", "--outdir", vendor_path, "--wheel"], check=True, cwd=union_root_path)

    if pull_flytekit_local_dev:
        expected_wheels += 1
        flytekit_root_path = Path(flytekit.__path__[0]).parent
        assert (flytekit_root_path / "pyproject.toml").exists(), "pyproject.toml needs to exists in flytekit directory"
        run([sys.executable, "-m", "build", "--outdir", vendor_path, "--wheel"], check=True, cwd=flytekit_root_path)

    wheels = list(vendor_path.iterdir())
    assert len(wheels) == expected_wheels, f"There should be {expected_wheels} wheel in {vendor_path}"
    packages = packages or []

    source_root_absolute = source_root.absolute()

    kwargs["source_copy_mode"] = CopyFileDetection.ALL

    if builder == "default":
        all_packages = " ".join(f"/root/.vendor/{wheel.name}" for wheel in wheels)
        if "commands" not in kwargs:
            kwargs["commands"] = []

        kwargs["commands"].insert(
            0,
            "--mount=from=uv,source=/uv,target=/usr/bin/uv uv pip install "
            f"--python /opt/micromamba/envs/runtime/bin/python {all_packages}",
        )
    elif builder == "envd":
        all_packages = " ".join(f"/root/.vendor/{wheel.name}" for wheel in wheels)
        if "commands" not in kwargs:
            kwargs["commands"] = []

        kwargs["commands"].insert(0, f"pip install {all_packages}")

    elif builder in ("unionai", "union"):
        for wheel in wheels:
            packages.append(f".vendor/{wheel.name}")
    else:
        raise ValueError("only default and union/unionai builder is supported")

    return ImageSpec(*args, builder=builder, packages=packages, source_root=os.fspath(source_root_absolute), **kwargs)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: objectstore/metadata_service.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from union.internal.objectstore import payload_pb2 as objectstore_dot_payload__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\"objectstore/metadata_service.proto\x12\x17\x63loudidl.objectstore.v1\x1a\x19objectstore/payload.proto2\x8c\x05\n\x14MetadataStoreService\x12\x61\n\x08Metadata\x12(.cloudidl.objectstore.v1.MetadataRequest\x1a).cloudidl.objectstore.v1.MetadataResponse\"\x00\x12U\n\x04Head\x12$.cloudidl.objectstore.v1.HeadRequest\x1a%.cloudidl.objectstore.v1.HeadResponse\"\x00\x12R\n\x03Put\x12#.cloudidl.objectstore.v1.PutRequest\x1a$.cloudidl.objectstore.v1.PutResponse\"\x00\x12R\n\x03Get\x12#.cloudidl.objectstore.v1.GetRequest\x1a$.cloudidl.objectstore.v1.GetResponse\"\x00\x12[\n\x06\x44\x65lete\x12&.cloudidl.objectstore.v1.DeleteRequest\x1a\'.cloudidl.objectstore.v1.DeleteResponse\"\x00\x12U\n\x04List\x12$.cloudidl.objectstore.v1.ListRequest\x1a%.cloudidl.objectstore.v1.ListResponse\"\x00\x12^\n\x07Presign\x12\'.cloudidl.objectstore.v1.PresignRequest\x1a(.cloudidl.objectstore.v1.PresignResponse\"\x00\x42\xe3\x01\n\x1b\x63om.cloudidl.objectstore.v1B\x14MetadataServiceProtoH\x02P\x01Z.github.com/unionai/cloud/gen/pb-go/objectstore\xa2\x02\x03\x43OX\xaa\x02\x17\x43loudidl.Objectstore.V1\xca\x02\x17\x43loudidl\\Objectstore\\V1\xe2\x02#Cloudidl\\Objectstore\\V1\\GPBMetadata\xea\x02\x19\x43loudidl::Objectstore::V1b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'objectstore.metadata_service_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\033com.cloudidl.objectstore.v1B\024MetadataServiceProtoH\002P\001Z.github.com/unionai/cloud/gen/pb-go/objectstore\242\002\003COX\252\002\027Cloudidl.Objectstore.V1\312\002\027Cloudidl\\Objectstore\\V1\342\002#Cloudidl\\Objectstore\\V1\\GPBMetadata\352\002\031Cloudidl::Objectstore::V1'
  _globals['_METADATASTORESERVICE']._serialized_start=91
  _globals['_METADATASTORESERVICE']._serialized_end=743
# @@protoc_insertion_point(module_scope)



import asyncio
import inspect
import json
import os
import platform
import shutil
import subprocess
import sys
import tarfile
from multiprocessing import Process
from pathlib import Path
from shutil import which
from string import Template
from subprocess import run
from textwrap import dedent
from time import time
from typing import List, NamedTuple, Optional, Union

import fsspec
from flytekit import PythonFunctionTask, current_context
from flytekit.configuration import AuthType, SerializationSettings
from flytekit.core.base_task import Task
from flytekit.core.context_manager import FlyteContextManager
from flytekit.core.utils import ClassDecorator
from flytekit.extend import TaskResolverMixin
from flytekit.types.directory import FlyteDirectory

import union
from union._async import merge, run_sync
from union._config import _VANITY_UNION_URLS, _clean_endpoint
from union.workspace._config import WorkspaceConfig
from union.workspace._vscode_remote import WorkspaceRemote

VSCODE_TYPE_VALUE = "vscode"
# Config keys to store in task template
TASK_FUNCTION_SOURCE_PATH = "TASK_FUNCTION_SOURCE_PATH"

WORKSPACE_IMAGE_NAME = "workspace"
CODE_SERVER_CLI_NAME = "code-server"
DEFAULT_PORT = 8080
DOWNLOAD_DIR = Path.home() / ".code-server"
WORKSPACE_DIR = Path("/workspace")
HEARTBEAT_PATH = Path.home() / ".local" / "share" / "code-server" / "heartbeat"
CODE_SERVER_SETTINGS_PATH = Path.home() / ".local" / "share" / "code-server" / "User" / "settings.json"
CODE_SERVER_VERSION = "4.23.1"
CODE_SERVER_DOWNLOAD_URL = (
    "https://github.com/coder/code-server/releases/download/v{version}/code-server-{version}-linux-{arch}.tar.gz"
)
CODE_SERVER_CLI_NAME_TEMPLATE = "code-server-{version}-linux-{arch}"
HEARTBEAT_CHECK_SECONDS = 60
WATCH_WORKSPACE_INSTANCE_CHECK_SECONDS = 30
GH_VERSION = "2.49.0"

ROOT_PATH = Path.home()
BASHRC_PATH = Path.home() / ".bashrc"
UNION_GITHUB_TOKEN_ENV = "UNION_WORKSPACE_GITHUB_TOKEN"
UNION_SERVERLESS_API_KEY_ENV = "UNION_WORKSPACE_SERVERLESS_API_KEY"
GIT_CONFIG_PATH = Path.home() / ".gitconfig"
TERMINATE_WORKSPACE_FILE_NAME = "terminate_workspace.py"
WORKSPACE_API_KEY_NAME = "WORKSPACE_API_KEY"


GIT_CONFIG_TEMPLATE = Template(
    """\
$USER_INFO

[credential "https://github.com"]
    helper =
    helper = !/usr/bin/gh auth git-credential
[credential "https://gist.github.com"]
    helper =
    helper = !/usr/bin/gh auth git-credential

[url "https://$GITHUB_TOKEN@github.com/"]
    insteadOf = https://github.com/
"""
)

WorkspaceOutput = NamedTuple("WorkspaceOutput", [("state_dir", FlyteDirectory)])


def configure_python_path():
    python_bin = os.path.dirname(sys.executable)
    with BASHRC_PATH.open("a") as f:
        f.write(os.linesep)
        f.write(f"export PATH={python_bin}:$PATH")
        f.write(os.linesep)
        f.write(r'export PS1="\[\e[0;34m\]\w\[\e[m\]\$ "')


def configure_union_remote(host: str):
    from union._config import _write_config_to_path

    _write_config_to_path(host, AuthType.DEVICEFLOW.value)


def configure_git(repo: str, workspace_dir: Path, git_name: str, git_email: str) -> str:
    """Configures workspace built with ImageSpec on Union's hosted image builder."""

    ctx = current_context()
    secrets = ctx.secrets

    with BASHRC_PATH.open("a") as f:
        github_token = secrets.get(key=UNION_GITHUB_TOKEN_ENV)
        f.write(f"export GITHUB_TOKEN={github_token}")

    user_info = dedent(
        f"""\
    [user]
        name = {git_name}
        email = {git_email}
    """
    )
    git_config = GIT_CONFIG_TEMPLATE.substitute(
        USER_INFO=user_info,
        GITHUB_TOKEN=github_token,
    )

    GIT_CONFIG_PATH.write_text(git_config)

    root_dir = workspace_dir

    if repo != "":
        workspace_dir.mkdir(exist_ok=True)
        subprocess.run(["/usr/bin/git", "clone", repo], cwd=workspace_dir, text=True)
        for item in workspace_dir.iterdir():
            if item.is_dir():
                root_dir = item
                break
    return root_dir


class vscode(ClassDecorator):
    """Union specific VSCode extension."""

    def __init__(self, task_function: Optional[callable] = None):
        super().__init__(task_function=task_function)

    def execute(self, *args, **kwargs):
        ctx = FlyteContextManager.current_context()
        ctx.user_space_params.builder().add_attr(
            TASK_FUNCTION_SOURCE_PATH, inspect.getsourcefile(self.task_function)
        ).build()

        if ctx.execution_state.is_local_execution():
            return

        return self.task_function(*args, **kwargs)

    def get_extra_config(self):
        return {self.LINK_TYPE_KEY: VSCODE_TYPE_VALUE, self.PORT_KEY: f"{DEFAULT_PORT}"}


def run_on_startup_commands(on_startup: Union[str, List[str]], workspace_dir: Path):
    on_startup = on_startup
    if on_startup is not None:
        if isinstance(on_startup, str):
            on_startup = [on_startup]

        commands = [
            # ensure that contents in the workspace directory is writable by the
            # user. This is needed e.g. when cloned git repos are checked out
            # as part of on_startup, but the state_dir needs to overwrite files
            # in the .git directory
            "chmod 775 -R .",
            *on_startup,
        ]

        for cmd in commands:
            subprocess.run(cmd.split(), cwd=workspace_dir, text=True)


@vscode
def workspace(
    config_content: str,
    host: str,
    console_http_domain: str,
    state_dir: Optional[FlyteDirectory] = None,
) -> WorkspaceOutput:
    ROOT_PATH.mkdir(exist_ok=True)
    configure_python_path()
    configure_union_remote(host)

    workspace_config = WorkspaceConfig.from_yaml(config_content)
    workspace_dir = Path(workspace_config.workspace_root).expanduser()
    workspace_dir.mkdir(exist_ok=True)

    code_server_cli = download_and_configure_vscode(CODE_SERVER_VERSION)
    run_on_startup_commands(workspace_config.on_startup, workspace_dir)

    if state_dir is not None:
        state_dir.download()
        shutil.copytree(state_dir.path, workspace_dir, dirs_exist_ok=True)

    # overwrite the config with the new config
    config_path = workspace_dir / "config.yaml"
    config_path.write_text(config_content)

    # Configure code-server
    configure_code_server_config()

    start_vscode_service(
        code_server_cli,
        host,
        console_http_domain,
        workspace_dir,
        workspace_config.ttl_seconds,
        workspace_config.working_dir,
    )
    logger = union.current_context().logging
    logger.info(f"Vscode service terminated, saving contents of {workspace_dir} to state_dir output")
    return WorkspaceOutput(state_dir=FlyteDirectory(path=str(workspace_dir)))


@union.task
def _workspace_image_build_noop(): ...


class WorkspaceResolver(TaskResolverMixin):
    @property
    def location(self) -> str:
        return "union.workspace._vscode.resolver"

    @property
    def name(self) -> str:
        return "union.workspace.vscode"

    def get_task_for_workspace(
        self,
        config: WorkspaceConfig,
    ) -> PythonFunctionTask:
        return PythonFunctionTask(
            task_config=None,
            task_function=_workspace_image_build_noop,
            task_type="_workspace_image_build_noop",
            task_resolver=self,
            requests=config.resources,
            container_image=config.container_image,
        )

    def load_task(self, loader_args: List[str]) -> PythonFunctionTask:
        return PythonFunctionTask(
            task_config=None,
            task_function=_workspace_image_build_noop,
            task_type="_workspace_image_build_noop",
            task_resolver=self,
        )

    def loader_args(self, settings: SerializationSettings, task: PythonFunctionTask) -> List[str]:
        return ["workspace"]

    def get_all_tasks(self) -> List[Task]:
        raise NotImplementedError


resolver = WorkspaceResolver()


def download_http(url: str, local_dest_path: Path) -> Path:
    """Download URL to `download_dir`. Returns Path of downloaded file."""
    logger = union.current_context().logging

    fs = fsspec.filesystem("http")
    logger.info(f"Downloading {url} to {local_dest_path}")
    fs.get(url, local_dest_path)
    logger.info("File downloaded successfully!")

    return local_dest_path


def download_and_configure_vscode(version: str) -> str:
    """Download and configure vscode."""
    logger = union.current_context().logging
    code_server_cli = which(CODE_SERVER_CLI_NAME)
    if code_server_cli is not None:
        logger.info(f"Code server binary already exists at {code_server_cli}")
        logger.info("Skipping downloading coe server")
        return code_server_cli

    # Download code-server
    logger.info(f"Code server is not in $PATH, downloading code server to {DOWNLOAD_DIR}...")
    DOWNLOAD_DIR.mkdir(exist_ok=True)

    machine_info = platform.machine()
    logger.info(f"machine type: {machine_info}")

    if machine_info == "aarch64":
        arch = "arm64"
    elif machine_info == "x86_64":
        arch = "amd64"
    else:
        msg = (
            "Automatic download is only supported on AMD64 and ARM64 architectures. "
            "If you are using a different architecture, please visit the code-server official "
            "website to manually download the appropriate version for your image."
        )
        raise ValueError(msg)

    code_server_tar_path_url = CODE_SERVER_DOWNLOAD_URL.format(version=version, arch=arch)
    code_server_local_path = DOWNLOAD_DIR / "code-server.tar.gz"
    download_http(code_server_tar_path_url, code_server_local_path)
    with tarfile.open(code_server_local_path, "r:gz") as tar:
        tar.extractall(path=DOWNLOAD_DIR)

    code_server_cli = (
        DOWNLOAD_DIR / CODE_SERVER_CLI_NAME_TEMPLATE.format(version=version, arch=arch) / "bin" / CODE_SERVER_CLI_NAME
    )
    return os.fspath(code_server_cli)


def configure_code_server_config():
    logger = union.current_context().logging
    config = {
        "terminal.integrated.defaultProfile.linux": "bash",
        "remote.autoForwardPortsSource": "hybrid",
        "commands.treeViewStatusBarVisibleSymbol": "",
        "commands.commands": {
            "⏹️ Terminate Workspace": {
                "sequence": [
                    {
                        "command": "commands.runInTerminal",
                        "args": {
                            "text": "python ~/workspace/terminate_workspace.py",
                            "name": "terminate",
                            "waitForExit": False,
                            "reuse": "newest",
                        },
                    },
                    {
                        "command": "commands.focusTerminal",
                        "args": "terminate",
                    },
                ],
                "statusBar": {
                    "alignment": "left",
                    "text": "$(stop) Terminate Workspace",
                    "backgroundColor": "warning",
                    "priority": -9999,
                },
            },
        },
        "workbench.colorCustomizations": {
            "sideBar.background": "#FFF8F3",
            "sideBarSectionHeader.background": "#FFEED5",
            "activityBar.background": "#FFF8F3",
            "statusBar.background": "#FFF8F3",
            "titleBar.activeBackground": "#FFF8F3",
        },
        "files.exclude": {
            "**/__pycache__/**": True,
            "**/.vscode/**": True,
        },
    }
    if not CODE_SERVER_SETTINGS_PATH.parent.exists():
        logger.info(f"Creating code server settings directory {CODE_SERVER_SETTINGS_PATH.parent}")
        CODE_SERVER_SETTINGS_PATH.parent.mkdir(parents=True, exist_ok=True)

    with CODE_SERVER_SETTINGS_PATH.open("w") as file:
        json.dump(config, file, indent=4)


def prepare_terminate_workspace_python(
    console_http_domain: str,
    workspace_root: Path,
    pid: int,
):
    """
    Generate a Python script for users to resume the task.
    """

    if console_http_domain.startswith("https://"):
        redirect_url = f"{console_http_domain}/"
    else:
        redirect_url = f"https://{console_http_domain}/"

    python_script = dedent(f"""\
    import argparse
    import os
    import signal
    import time
    import webbrowser

    if __name__ == "__main__":
        parser = argparse.ArgumentParser()
        parser.add_argument("--force", action="store_true", help="Force terminate the workspace.")
        args = parser.parse_args()

        if args.force:
            answer = "Y"
        else:
            print(
                "⛔️ This operation will terminate the workspace. The contents "
                "in the {workspace_root} directory will not be saved."
            )
            answer = input("Do you really want to terminate? [Y/n]: ").strip().upper()

        if answer in ['Y', '']:
            os.kill({pid}, signal.SIGTERM)
            print(f"🛑 The server is terminating. You will be redirected to Union in a few seconds.")
            time.sleep(3)
            webbrowser.open("{redirect_url}", new=0)
            print("⤴️ Redirecting to Union.")
        else:
            print("Operation canceled.")
    """)

    with (workspace_root / TERMINATE_WORKSPACE_FILE_NAME).open("w") as file:
        file.write(python_script)


def prepare_task_json(workspace_root: Path, working_dir: Path):
    tasks_json = {
        "version": "2.0.0",
        "tasks": [
            {
                "label": "terminate",
                "type": "shell",
                "command": f"python {str(workspace_root / TERMINATE_WORKSPACE_FILE_NAME)}",
            }
        ],
    }

    vscode_directory = working_dir / ".vscode"
    if not os.path.exists(vscode_directory):
        os.makedirs(vscode_directory)

    with open(os.path.join(vscode_directory, "tasks.json"), "w") as file:
        json.dump(tasks_json, file, indent=4)


def start_vscode_service(
    code_server_cli: str,
    host: str,
    console_http_domain: str,
    workspace_root: Path,
    ttl_seconds: int,
    working_dir: Optional[str] = None,
):
    logger = union.current_context().logging
    code_server_host = f"0.0.0.0:{DEFAULT_PORT}"

    command = [
        code_server_cli,
        "--bind-addr",
        code_server_host,
        "--disable-workspace-trust",
        "--auth",
        "none",
    ]

    workspace_root = workspace_root.absolute()
    if working_dir is not None:
        working_dir: Path = workspace_root / working_dir
    else:
        working_dir: Path = workspace_root

    command.append(os.fspath(working_dir))

    if not workspace_root.exists():
        logger.info(f"Creating workspace root directory {workspace_root}")
        workspace_root.mkdir(parents=True, exist_ok=True)

    if not working_dir.exists():
        logger.info(f"Creating working directory {working_dir}")
        working_dir.mkdir(parents=True, exist_ok=True)

    workspace_api_key = f"{os.getenv('FLYTE_SECRETS_ENV_PREFIX')}{WORKSPACE_API_KEY_NAME}"
    if workspace_api_key in os.environ:
        if "UNION_API_KEY" not in os.environ:
            os.environ["UNION_API_KEY"] = os.environ[workspace_api_key]
    else:
        raise RuntimeError(f"{workspace_api_key} is not set")

    env = os.environ.copy()
    host = _clean_endpoint(host)
    env["_UNION_ENDPOINT"] = _VANITY_UNION_URLS.get(host, host)

    child_process = Process(
        target=run,
        args=[command],
        kwargs={"text": True, "env": env},
    )
    child_process.start()

    logger.info(f"Preparing terminate workspace python script in {workspace_root}, pid: {child_process.pid}")
    prepare_terminate_workspace_python(console_http_domain, workspace_root, child_process.pid)

    logger.info(f"Preparing launch.json in {workspace_root}")
    prepare_task_json(workspace_root, working_dir)

    project = os.getenv("FLYTE_INTERNAL_PROJECT")
    domain = os.getenv("FLYTE_INTERNAL_DOMAIN")

    assert project is not None, "FLYTE_INTERNAL_PROJECT must be set"
    assert domain is not None, "FLYTE_INTERNAL_DOMAIN must be set"

    workspace_instance_name = os.getenv("UNION_INTERNAL_WORKSPACE_INSTANCE_NAME")
    assert workspace_instance_name is not None, "UNION_INTERNAL_WORKSPACE_INSTANCE_NAME must be set"

    remote = WorkspaceRemote(default_project=project, default_domain=domain, union_remote=union.UnionRemote())
    run_sync(
        monitor_vscode,
        remote,
        workspace_instance_name,
        child_process,
        ttl_seconds,
    )
    return


async def watch_heartbeat(
    child_process: Process,
    start_time: float,
    ttl_seconds: int,
    logger,
):
    while child_process.is_alive():
        if not HEARTBEAT_PATH.exists():
            logger.info(f"Code server heartbeat file does not exist, using start time {start_time}")
            delta = time() - start_time
        else:
            file_mod_time = HEARTBEAT_PATH.stat().st_mtime
            logger.info(f"Code server heartbeat file exists, using file modification time {file_mod_time}")
            delta = time() - file_mod_time

        logger.info(f"The latest activity on code server was {delta} ago.")
        if delta > ttl_seconds:
            logger.info(
                f"code server is idle for more than {ttl_seconds} seconds. Terminating process {child_process.pid}"
            )
            yield True
            break

        logger.info("Continuing to monitor code server heartbeat")
        yield False
        await asyncio.sleep(HEARTBEAT_CHECK_SECONDS)
    else:
        logger.info("Code server process is already terminated")
        yield True


async def poll_workspace_instance(
    child_process: Process,
    remote: WorkspaceRemote,
    workspace_instance_name: str,
    logger,
):
    """
    This is a fallback method to make sure that the workspace instance is stopping.
    """
    while child_process.is_alive():
        logger.info("Waiting for workspace instance to stop")
        is_stopping = await remote.workspace_instance_is_stopping(workspace_instance_name, logger=logger)
        if is_stopping:
            logger.info("Workspace instance is stopping")
            yield is_stopping
            break

        logger.info("Sleeping workspace instance check")
        yield False
        await asyncio.sleep(WATCH_WORKSPACE_INSTANCE_CHECK_SECONDS)
    else:
        logger.info("Code server process is already terminated")
        yield True


async def monitor_vscode(
    remote: WorkspaceRemote,
    workspace_instance_name: str,
    child_process: Process,
    ttl_seconds: int,
):
    logger = union.current_context().logging
    start_time = time()

    loop = asyncio.get_running_loop()

    logger.info("Starting to monitor workspace instance for termination signal")

    watch_heartbeat_task = watch_heartbeat(child_process, start_time, ttl_seconds, logger)
    watch_workspace_instance_task = remote.watch_workspace_instance(workspace_instance_name, logger=logger)
    poll_workspace_instance_task = poll_workspace_instance(child_process, remote, workspace_instance_name, logger)

    try:
        async for stop_event in merge(
            loop,
            watch_heartbeat_task,
            # watch_workspace_instance_task,
            poll_workspace_instance_task,
        ):
            if stop_event:
                child_process.terminate()
                child_process.join()
                logger.info(f"Code server process {child_process.pid} terminated")
                return
            else:
                logger.info("Code server process is still running")
    finally:
        watch_heartbeat_task.aclose()
        watch_workspace_instance_task.aclose()
        poll_workspace_instance_task.aclose()



from union.internal.validate.validate import validate_pb2 as _validate_pb2
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class ProjectIdentifier(_message.Message):
    __slots__ = ["organization", "domain", "name"]
    ORGANIZATION_FIELD_NUMBER: _ClassVar[int]
    DOMAIN_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    organization: str
    domain: str
    name: str
    def __init__(self, organization: _Optional[str] = ..., domain: _Optional[str] = ..., name: _Optional[str] = ...) -> None: ...

class ClusterIdentifier(_message.Message):
    __slots__ = ["organization", "name"]
    ORGANIZATION_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    organization: str
    name: str
    def __init__(self, organization: _Optional[str] = ..., name: _Optional[str] = ...) -> None: ...

class ClusterPoolIdentifier(_message.Message):
    __slots__ = ["organization", "name"]
    ORGANIZATION_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    organization: str
    name: str
    def __init__(self, organization: _Optional[str] = ..., name: _Optional[str] = ...) -> None: ...

class ClusterConfigIdentifier(_message.Message):
    __slots__ = ["organization", "id"]
    ORGANIZATION_FIELD_NUMBER: _ClassVar[int]
    ID_FIELD_NUMBER: _ClassVar[int]
    organization: str
    id: str
    def __init__(self, organization: _Optional[str] = ..., id: _Optional[str] = ...) -> None: ...

class UserIdentifier(_message.Message):
    __slots__ = ["subject"]
    SUBJECT_FIELD_NUMBER: _ClassVar[int]
    subject: str
    def __init__(self, subject: _Optional[str] = ...) -> None: ...

class ApplicationIdentifier(_message.Message):
    __slots__ = ["subject"]
    SUBJECT_FIELD_NUMBER: _ClassVar[int]
    subject: str
    def __init__(self, subject: _Optional[str] = ...) -> None: ...

class RoleIdentifier(_message.Message):
    __slots__ = ["organization", "name"]
    ORGANIZATION_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    organization: str
    name: str
    def __init__(self, organization: _Optional[str] = ..., name: _Optional[str] = ...) -> None: ...

class OrgIdentifier(_message.Message):
    __slots__ = ["name"]
    NAME_FIELD_NUMBER: _ClassVar[int]
    name: str
    def __init__(self, name: _Optional[str] = ...) -> None: ...

class ManagedClusterIdentifier(_message.Message):
    __slots__ = ["name", "org"]
    NAME_FIELD_NUMBER: _ClassVar[int]
    ORG_FIELD_NUMBER: _ClassVar[int]
    name: str
    org: OrgIdentifier
    def __init__(self, name: _Optional[str] = ..., org: _Optional[_Union[OrgIdentifier, _Mapping]] = ...) -> None: ...

class PolicyIdentifier(_message.Message):
    __slots__ = ["organization", "name"]
    ORGANIZATION_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    organization: str
    name: str
    def __init__(self, organization: _Optional[str] = ..., name: _Optional[str] = ...) -> None: ...



import os
from dataclasses import dataclass
from typing import List, Optional, Union

from flytekit import ImageSpec, Resources
from mashumaro.mixins.json import DataClassJSONMixin
from mashumaro.mixins.yaml import DataClassYAMLMixin

WORKSPACE_IMAGE_NAME = "workspace"


class ImageSpecConfig(DataClassYAMLMixin, DataClassJSONMixin, ImageSpec):
    def __hash__(self):
        return hash(self.to_dict().__str__())


class ResourceConfig(DataClassYAMLMixin, Resources):
    pass


@dataclass
class WorkspaceConfig(DataClassYAMLMixin, DataClassJSONMixin):
    name: str
    project: str
    domain: str
    description: Optional[str] = None
    container_image: Optional[Union[ImageSpecConfig, str]] = None
    resources: Optional[ResourceConfig] = None
    accelerator: Optional[str] = None
    on_startup: Optional[Union[str, List[str]]] = None
    workspace_root: str = "~/workspace"
    working_dir: Optional[str] = None
    ttl_seconds: int = 1200  # 20 minutes is the default up time

    def __post_init__(self):
        # Quick hack to always build image
        if isinstance(self.container_image, ImageSpecConfig):
            from union.ucimage._image_builder import _register_union_image_builder

            _register_union_image_builder()

            if os.getenv("UNION_DEV"):
                from union._testing import imagespec_with_local_union

                image_spec = imagespec_with_local_union(builder="union")
            else:
                image_spec = ImageSpec(builder="union")

            _user_image_spec = self.container_image
            self.container_image = ImageSpecConfig(
                name=f"{WORKSPACE_IMAGE_NAME}-{self.container_image.name}",
                builder=image_spec.builder,
                packages=[
                    *(image_spec.packages or []),
                    *(_user_image_spec.packages or []),
                ],
                source_root=image_spec.source_root,
                apt_packages=[
                    *(_user_image_spec.apt_packages or []),
                    "git",
                    "sudo",
                    "wget",
                    "vim",
                ],
                commands=[
                    "wget https://github.com/cli/cli/releases/download/v2.49.0/gh_2.49.0_linux_amd64.deb "
                    "-O /tmp/gh.deb",
                    "apt install /tmp/gh.deb",
                    "wget https://github.com/coder/code-server/releases/download/v4.23.1/code-server_4.23.1_amd64.deb "
                    "-O /tmp/code-server.deb",
                    "apt install /tmp/code-server.deb",
                    *(_user_image_spec.commands or []),
                ],
            )


_DEFAULT_CONFIG_YAML_FOR_BASE_IMAGE = """\
name: my-workspace
description: my workspace
# Make sure that the project and domain exists
project: {default_project}
domain: development
container_image: public.ecr.aws/unionai/workspace-base:py3.11-latest
resources:
    cpu: "2"
    mem: "4Gi"
    gpu: null
accelerator: null
on_startup: null
ttl_seconds: 1200
"""

_DEFAULT_CONFIG_YAML_FOR_IMAGE_SPEC = """\
name: my-workspace
description: my workspace
project: {default_project}
domain: development
container_image:
    builder: "union"
    packages:
    - torch
resources:
    cpu: "2"
    mem: "4Gi"
    gpu: null
accelerator: null
on_startup: null
ttl_seconds: 1200
"""



from union.internal.common import identity_pb2 as _identity_pb2
from union.internal.common import runtime_version_pb2 as _runtime_version_pb2
from flyteidl.core import artifact_id_pb2 as _artifact_id_pb2
from flyteidl.core import security_pb2 as _security_pb2
from flyteidl.core import tasks_pb2 as _tasks_pb2
from google.protobuf import duration_pb2 as _duration_pb2
from google.protobuf import timestamp_pb2 as _timestamp_pb2
from union.internal.validate.validate import validate_pb2 as _validate_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class Identifier(_message.Message):
    __slots__ = ["org", "project", "domain", "name"]
    ORG_FIELD_NUMBER: _ClassVar[int]
    PROJECT_FIELD_NUMBER: _ClassVar[int]
    DOMAIN_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    org: str
    project: str
    domain: str
    name: str
    def __init__(self, org: _Optional[str] = ..., project: _Optional[str] = ..., domain: _Optional[str] = ..., name: _Optional[str] = ...) -> None: ...

class Meta(_message.Message):
    __slots__ = ["id", "revision", "labels"]
    class LabelsEntry(_message.Message):
        __slots__ = ["key", "value"]
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
    ID_FIELD_NUMBER: _ClassVar[int]
    REVISION_FIELD_NUMBER: _ClassVar[int]
    LABELS_FIELD_NUMBER: _ClassVar[int]
    id: Identifier
    revision: int
    labels: _containers.ScalarMap[str, str]
    def __init__(self, id: _Optional[_Union[Identifier, _Mapping]] = ..., revision: _Optional[int] = ..., labels: _Optional[_Mapping[str, str]] = ...) -> None: ...

class AppWrapper(_message.Message):
    __slots__ = ["host", "app", "app_id"]
    HOST_FIELD_NUMBER: _ClassVar[int]
    APP_FIELD_NUMBER: _ClassVar[int]
    APP_ID_FIELD_NUMBER: _ClassVar[int]
    host: str
    app: App
    app_id: Identifier
    def __init__(self, host: _Optional[str] = ..., app: _Optional[_Union[App, _Mapping]] = ..., app_id: _Optional[_Union[Identifier, _Mapping]] = ...) -> None: ...

class App(_message.Message):
    __slots__ = ["metadata", "spec", "status"]
    METADATA_FIELD_NUMBER: _ClassVar[int]
    SPEC_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    metadata: Meta
    spec: Spec
    status: Status
    def __init__(self, metadata: _Optional[_Union[Meta, _Mapping]] = ..., spec: _Optional[_Union[Spec, _Mapping]] = ..., status: _Optional[_Union[Status, _Mapping]] = ...) -> None: ...

class Condition(_message.Message):
    __slots__ = ["last_transition_time", "deployment_status", "message", "revision", "actor"]
    LAST_TRANSITION_TIME_FIELD_NUMBER: _ClassVar[int]
    DEPLOYMENT_STATUS_FIELD_NUMBER: _ClassVar[int]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    REVISION_FIELD_NUMBER: _ClassVar[int]
    ACTOR_FIELD_NUMBER: _ClassVar[int]
    last_transition_time: _timestamp_pb2.Timestamp
    deployment_status: Status.DeploymentStatus
    message: str
    revision: int
    actor: _identity_pb2.EnrichedIdentity
    def __init__(self, last_transition_time: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., deployment_status: _Optional[_Union[Status.DeploymentStatus, str]] = ..., message: _Optional[str] = ..., revision: _Optional[int] = ..., actor: _Optional[_Union[_identity_pb2.EnrichedIdentity, _Mapping]] = ...) -> None: ...

class Status(_message.Message):
    __slots__ = ["assigned_cluster", "current_replicas", "ingress", "created_at", "last_updated_at", "conditions", "lease_expiration", "k8s_metadata", "materialized_inputs"]
    class DeploymentStatus(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = []
        DEPLOYMENT_STATUS_UNSPECIFIED: _ClassVar[Status.DeploymentStatus]
        DEPLOYMENT_STATUS_UNASSIGNED: _ClassVar[Status.DeploymentStatus]
        DEPLOYMENT_STATUS_ASSIGNED: _ClassVar[Status.DeploymentStatus]
        DEPLOYMENT_STATUS_PENDING: _ClassVar[Status.DeploymentStatus]
        DEPLOYMENT_STATUS_STOPPED: _ClassVar[Status.DeploymentStatus]
        DEPLOYMENT_STATUS_STARTED: _ClassVar[Status.DeploymentStatus]
        DEPLOYMENT_STATUS_FAILED: _ClassVar[Status.DeploymentStatus]
    DEPLOYMENT_STATUS_UNSPECIFIED: Status.DeploymentStatus
    DEPLOYMENT_STATUS_UNASSIGNED: Status.DeploymentStatus
    DEPLOYMENT_STATUS_ASSIGNED: Status.DeploymentStatus
    DEPLOYMENT_STATUS_PENDING: Status.DeploymentStatus
    DEPLOYMENT_STATUS_STOPPED: Status.DeploymentStatus
    DEPLOYMENT_STATUS_STARTED: Status.DeploymentStatus
    DEPLOYMENT_STATUS_FAILED: Status.DeploymentStatus
    ASSIGNED_CLUSTER_FIELD_NUMBER: _ClassVar[int]
    CURRENT_REPLICAS_FIELD_NUMBER: _ClassVar[int]
    INGRESS_FIELD_NUMBER: _ClassVar[int]
    CREATED_AT_FIELD_NUMBER: _ClassVar[int]
    LAST_UPDATED_AT_FIELD_NUMBER: _ClassVar[int]
    CONDITIONS_FIELD_NUMBER: _ClassVar[int]
    LEASE_EXPIRATION_FIELD_NUMBER: _ClassVar[int]
    K8S_METADATA_FIELD_NUMBER: _ClassVar[int]
    MATERIALIZED_INPUTS_FIELD_NUMBER: _ClassVar[int]
    assigned_cluster: str
    current_replicas: int
    ingress: Ingress
    created_at: _timestamp_pb2.Timestamp
    last_updated_at: _timestamp_pb2.Timestamp
    conditions: _containers.RepeatedCompositeFieldContainer[Condition]
    lease_expiration: _timestamp_pb2.Timestamp
    k8s_metadata: K8sMetadata
    materialized_inputs: MaterializedInputs
    def __init__(self, assigned_cluster: _Optional[str] = ..., current_replicas: _Optional[int] = ..., ingress: _Optional[_Union[Ingress, _Mapping]] = ..., created_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., last_updated_at: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., conditions: _Optional[_Iterable[_Union[Condition, _Mapping]]] = ..., lease_expiration: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., k8s_metadata: _Optional[_Union[K8sMetadata, _Mapping]] = ..., materialized_inputs: _Optional[_Union[MaterializedInputs, _Mapping]] = ...) -> None: ...

class K8sMetadata(_message.Message):
    __slots__ = ["namespace"]
    NAMESPACE_FIELD_NUMBER: _ClassVar[int]
    namespace: str
    def __init__(self, namespace: _Optional[str] = ...) -> None: ...

class Ingress(_message.Message):
    __slots__ = ["public_url", "cname_url", "vpc_url"]
    PUBLIC_URL_FIELD_NUMBER: _ClassVar[int]
    CNAME_URL_FIELD_NUMBER: _ClassVar[int]
    VPC_URL_FIELD_NUMBER: _ClassVar[int]
    public_url: str
    cname_url: str
    vpc_url: str
    def __init__(self, public_url: _Optional[str] = ..., cname_url: _Optional[str] = ..., vpc_url: _Optional[str] = ...) -> None: ...

class Spec(_message.Message):
    __slots__ = ["container", "pod", "autoscaling", "ingress", "desired_state", "cluster_pool", "images", "security_context", "extended_resources", "runtime_metadata", "profile", "creator", "inputs"]
    class DesiredState(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = []
        DESIRED_STATE_UNSPECIFIED: _ClassVar[Spec.DesiredState]
        DESIRED_STATE_STOPPED: _ClassVar[Spec.DesiredState]
        DESIRED_STATE_STARTED: _ClassVar[Spec.DesiredState]
    DESIRED_STATE_UNSPECIFIED: Spec.DesiredState
    DESIRED_STATE_STOPPED: Spec.DesiredState
    DESIRED_STATE_STARTED: Spec.DesiredState
    CONTAINER_FIELD_NUMBER: _ClassVar[int]
    POD_FIELD_NUMBER: _ClassVar[int]
    AUTOSCALING_FIELD_NUMBER: _ClassVar[int]
    INGRESS_FIELD_NUMBER: _ClassVar[int]
    DESIRED_STATE_FIELD_NUMBER: _ClassVar[int]
    CLUSTER_POOL_FIELD_NUMBER: _ClassVar[int]
    IMAGES_FIELD_NUMBER: _ClassVar[int]
    SECURITY_CONTEXT_FIELD_NUMBER: _ClassVar[int]
    EXTENDED_RESOURCES_FIELD_NUMBER: _ClassVar[int]
    RUNTIME_METADATA_FIELD_NUMBER: _ClassVar[int]
    PROFILE_FIELD_NUMBER: _ClassVar[int]
    CREATOR_FIELD_NUMBER: _ClassVar[int]
    INPUTS_FIELD_NUMBER: _ClassVar[int]
    container: _tasks_pb2.Container
    pod: _tasks_pb2.K8sPod
    autoscaling: AutoscalingConfig
    ingress: IngressConfig
    desired_state: Spec.DesiredState
    cluster_pool: str
    images: ImageSpecSet
    security_context: SecurityContext
    extended_resources: _tasks_pb2.ExtendedResources
    runtime_metadata: _runtime_version_pb2.RuntimeMetadata
    profile: Profile
    creator: _identity_pb2.EnrichedIdentity
    inputs: InputList
    def __init__(self, container: _Optional[_Union[_tasks_pb2.Container, _Mapping]] = ..., pod: _Optional[_Union[_tasks_pb2.K8sPod, _Mapping]] = ..., autoscaling: _Optional[_Union[AutoscalingConfig, _Mapping]] = ..., ingress: _Optional[_Union[IngressConfig, _Mapping]] = ..., desired_state: _Optional[_Union[Spec.DesiredState, str]] = ..., cluster_pool: _Optional[str] = ..., images: _Optional[_Union[ImageSpecSet, _Mapping]] = ..., security_context: _Optional[_Union[SecurityContext, _Mapping]] = ..., extended_resources: _Optional[_Union[_tasks_pb2.ExtendedResources, _Mapping]] = ..., runtime_metadata: _Optional[_Union[_runtime_version_pb2.RuntimeMetadata, _Mapping]] = ..., profile: _Optional[_Union[Profile, _Mapping]] = ..., creator: _Optional[_Union[_identity_pb2.EnrichedIdentity, _Mapping]] = ..., inputs: _Optional[_Union[InputList, _Mapping]] = ...) -> None: ...

class Input(_message.Message):
    __slots__ = ["name", "string_value", "artifact_query", "artifact_id", "app_id"]
    NAME_FIELD_NUMBER: _ClassVar[int]
    STRING_VALUE_FIELD_NUMBER: _ClassVar[int]
    ARTIFACT_QUERY_FIELD_NUMBER: _ClassVar[int]
    ARTIFACT_ID_FIELD_NUMBER: _ClassVar[int]
    APP_ID_FIELD_NUMBER: _ClassVar[int]
    name: str
    string_value: str
    artifact_query: _artifact_id_pb2.ArtifactQuery
    artifact_id: _artifact_id_pb2.ArtifactID
    app_id: Identifier
    def __init__(self, name: _Optional[str] = ..., string_value: _Optional[str] = ..., artifact_query: _Optional[_Union[_artifact_id_pb2.ArtifactQuery, _Mapping]] = ..., artifact_id: _Optional[_Union[_artifact_id_pb2.ArtifactID, _Mapping]] = ..., app_id: _Optional[_Union[Identifier, _Mapping]] = ...) -> None: ...

class MaterializedInputs(_message.Message):
    __slots__ = ["items", "revision"]
    ITEMS_FIELD_NUMBER: _ClassVar[int]
    REVISION_FIELD_NUMBER: _ClassVar[int]
    items: _containers.RepeatedCompositeFieldContainer[MaterializedInput]
    revision: int
    def __init__(self, items: _Optional[_Iterable[_Union[MaterializedInput, _Mapping]]] = ..., revision: _Optional[int] = ...) -> None: ...

class MaterializedInput(_message.Message):
    __slots__ = ["name", "artifact_id"]
    NAME_FIELD_NUMBER: _ClassVar[int]
    ARTIFACT_ID_FIELD_NUMBER: _ClassVar[int]
    name: str
    artifact_id: _artifact_id_pb2.ArtifactID
    def __init__(self, name: _Optional[str] = ..., artifact_id: _Optional[_Union[_artifact_id_pb2.ArtifactID, _Mapping]] = ...) -> None: ...

class InputList(_message.Message):
    __slots__ = ["items"]
    ITEMS_FIELD_NUMBER: _ClassVar[int]
    items: _containers.RepeatedCompositeFieldContainer[Input]
    def __init__(self, items: _Optional[_Iterable[_Union[Input, _Mapping]]] = ...) -> None: ...

class Profile(_message.Message):
    __slots__ = ["type", "name", "short_description", "icon_url"]
    TYPE_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    SHORT_DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    ICON_URL_FIELD_NUMBER: _ClassVar[int]
    type: str
    name: str
    short_description: str
    icon_url: str
    def __init__(self, type: _Optional[str] = ..., name: _Optional[str] = ..., short_description: _Optional[str] = ..., icon_url: _Optional[str] = ...) -> None: ...

class SecurityContext(_message.Message):
    __slots__ = ["run_as", "secrets", "allow_anonymous"]
    RUN_AS_FIELD_NUMBER: _ClassVar[int]
    SECRETS_FIELD_NUMBER: _ClassVar[int]
    ALLOW_ANONYMOUS_FIELD_NUMBER: _ClassVar[int]
    run_as: _security_pb2.Identity
    secrets: _containers.RepeatedCompositeFieldContainer[_security_pb2.Secret]
    allow_anonymous: bool
    def __init__(self, run_as: _Optional[_Union[_security_pb2.Identity, _Mapping]] = ..., secrets: _Optional[_Iterable[_Union[_security_pb2.Secret, _Mapping]]] = ..., allow_anonymous: bool = ...) -> None: ...

class ImageSpec(_message.Message):
    __slots__ = ["tag", "build_job_url"]
    TAG_FIELD_NUMBER: _ClassVar[int]
    BUILD_JOB_URL_FIELD_NUMBER: _ClassVar[int]
    tag: str
    build_job_url: str
    def __init__(self, tag: _Optional[str] = ..., build_job_url: _Optional[str] = ...) -> None: ...

class ImageSpecSet(_message.Message):
    __slots__ = ["images"]
    IMAGES_FIELD_NUMBER: _ClassVar[int]
    images: _containers.RepeatedCompositeFieldContainer[ImageSpec]
    def __init__(self, images: _Optional[_Iterable[_Union[ImageSpec, _Mapping]]] = ...) -> None: ...

class IngressConfig(_message.Message):
    __slots__ = ["private", "subdomain", "cname"]
    PRIVATE_FIELD_NUMBER: _ClassVar[int]
    SUBDOMAIN_FIELD_NUMBER: _ClassVar[int]
    CNAME_FIELD_NUMBER: _ClassVar[int]
    private: bool
    subdomain: str
    cname: str
    def __init__(self, private: bool = ..., subdomain: _Optional[str] = ..., cname: _Optional[str] = ...) -> None: ...

class AutoscalingConfig(_message.Message):
    __slots__ = ["replicas", "scaledown_period", "scaling_metric"]
    REPLICAS_FIELD_NUMBER: _ClassVar[int]
    SCALEDOWN_PERIOD_FIELD_NUMBER: _ClassVar[int]
    SCALING_METRIC_FIELD_NUMBER: _ClassVar[int]
    replicas: Replicas
    scaledown_period: _duration_pb2.Duration
    scaling_metric: ScalingMetric
    def __init__(self, replicas: _Optional[_Union[Replicas, _Mapping]] = ..., scaledown_period: _Optional[_Union[_duration_pb2.Duration, _Mapping]] = ..., scaling_metric: _Optional[_Union[ScalingMetric, _Mapping]] = ...) -> None: ...

class ScalingMetric(_message.Message):
    __slots__ = ["request_rate", "concurrency"]
    REQUEST_RATE_FIELD_NUMBER: _ClassVar[int]
    CONCURRENCY_FIELD_NUMBER: _ClassVar[int]
    request_rate: RequestRate
    concurrency: Concurrency
    def __init__(self, request_rate: _Optional[_Union[RequestRate, _Mapping]] = ..., concurrency: _Optional[_Union[Concurrency, _Mapping]] = ...) -> None: ...

class Concurrency(_message.Message):
    __slots__ = ["target_value"]
    TARGET_VALUE_FIELD_NUMBER: _ClassVar[int]
    target_value: int
    def __init__(self, target_value: _Optional[int] = ...) -> None: ...

class RequestRate(_message.Message):
    __slots__ = ["target_value"]
    TARGET_VALUE_FIELD_NUMBER: _ClassVar[int]
    target_value: int
    def __init__(self, target_value: _Optional[int] = ...) -> None: ...

class Replicas(_message.Message):
    __slots__ = ["min", "max"]
    MIN_FIELD_NUMBER: _ClassVar[int]
    MAX_FIELD_NUMBER: _ClassVar[int]
    min: int
    max: int
    def __init__(self, min: _Optional[int] = ..., max: _Optional[int] = ...) -> None: ...



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: workspace/workspace_definition_service.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.api import annotations_pb2 as google_dot_api_dot_annotations__pb2
from union.internal.workspace import workspace_definition_payload_pb2 as workspace_dot_workspace__definition__payload__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n,workspace/workspace_definition_service.proto\x12\x12\x63loudidl.workspace\x1a\x1cgoogle/api/annotations.proto\x1a,workspace/workspace_definition_payload.proto2\x92\x0b\n\x18WorkspaceRegistryService\x12\xaf\x01\n\x19\x43reateWorkspaceDefinition\x12\x34.cloudidl.workspace.CreateWorkspaceDefinitionRequest\x1a\x35.cloudidl.workspace.CreateWorkspaceDefinitionResponse\"%\x82\xd3\xe4\x93\x02\x1f\"\x1a/workspace_registry/api/v1:\x01*\x12\xdf\x01\n\x16GetWorkspaceDefinition\x12\x31.cloudidl.workspace.GetWorkspaceDefinitionRequest\x1a\x32.cloudidl.workspace.GetWorkspaceDefinitionResponse\"^\x90\x02\x01\x82\xd3\xe4\x93\x02U\x12S/workspace_registry/api/v1/{id.org}/{id.project}/{id.domain}/{id.name}/{id.version}\x12\x8d\x02\n\x1fUpdateWorkspaceDefinitionEntity\x12:.cloudidl.workspace.UpdateWorkspaceDefinitionEntityRequest\x1a;.cloudidl.workspace.UpdateWorkspaceDefinitionEntityResponse\"q\x82\xd3\xe4\x93\x02k\x1ai/workspace_registry/api/v1/entity/{entity.id.org}/{entity.id.project}/{entity.id.domain}/{entity.id.name}\x12\xe5\x01\n\x19UpdateWorkspaceDefinition\x12\x34.cloudidl.workspace.UpdateWorkspaceDefinitionRequest\x1a\x35.cloudidl.workspace.UpdateWorkspaceDefinitionResponse\"[\x82\xd3\xe4\x93\x02U\x1aS/workspace_registry/api/v1/{id.org}/{id.project}/{id.domain}/{id.name}/{id.version}\x12\xe3\x01\n\x1fListWorkspaceDefinitionEntities\x12:.cloudidl.workspace.ListWorkspaceDefinitionEntitiesRequest\x1a;.cloudidl.workspace.ListWorkspaceDefinitionEntitiesResponse\"G\x90\x02\x01\x82\xd3\xe4\x93\x02>\x12</workspace_registry/api/v1/entities/{org}/{project}/{domain}\x12\x83\x02\n\x18ListWorkspaceDefinitions\x12\x33.cloudidl.workspace.ListWorkspaceDefinitionsRequest\x1a\x34.cloudidl.workspace.ListWorkspaceDefinitionsResponse\"|\x90\x02\x01\x82\xd3\xe4\x93\x02s\x12\x33/workspace_registry/api/v1/{org}/{project}/{domain}Z<\x12:/workspace_registry/api/v1/{org}/{project}/{domain}/{name}B\xd2\x01\n\x16\x63om.cloudidl.workspaceB\x1fWorkspaceDefinitionServiceProtoH\x02P\x01Z,github.com/unionai/cloud/gen/pb-go/workspace\xa2\x02\x03\x43WX\xaa\x02\x12\x43loudidl.Workspace\xca\x02\x12\x43loudidl\\Workspace\xe2\x02\x1e\x43loudidl\\Workspace\\GPBMetadata\xea\x02\x13\x43loudidl::Workspaceb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'workspace.workspace_definition_service_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\026com.cloudidl.workspaceB\037WorkspaceDefinitionServiceProtoH\002P\001Z,github.com/unionai/cloud/gen/pb-go/workspace\242\002\003CWX\252\002\022Cloudidl.Workspace\312\002\022Cloudidl\\Workspace\342\002\036Cloudidl\\Workspace\\GPBMetadata\352\002\023Cloudidl::Workspace'
  _WORKSPACEREGISTRYSERVICE.methods_by_name['CreateWorkspaceDefinition']._options = None
  _WORKSPACEREGISTRYSERVICE.methods_by_name['CreateWorkspaceDefinition']._serialized_options = b'\202\323\344\223\002\037\"\032/workspace_registry/api/v1:\001*'
  _WORKSPACEREGISTRYSERVICE.methods_by_name['GetWorkspaceDefinition']._options = None
  _WORKSPACEREGISTRYSERVICE.methods_by_name['GetWorkspaceDefinition']._serialized_options = b'\220\002\001\202\323\344\223\002U\022S/workspace_registry/api/v1/{id.org}/{id.project}/{id.domain}/{id.name}/{id.version}'
  _WORKSPACEREGISTRYSERVICE.methods_by_name['UpdateWorkspaceDefinitionEntity']._options = None
  _WORKSPACEREGISTRYSERVICE.methods_by_name['UpdateWorkspaceDefinitionEntity']._serialized_options = b'\202\323\344\223\002k\032i/workspace_registry/api/v1/entity/{entity.id.org}/{entity.id.project}/{entity.id.domain}/{entity.id.name}'
  _WORKSPACEREGISTRYSERVICE.methods_by_name['UpdateWorkspaceDefinition']._options = None
  _WORKSPACEREGISTRYSERVICE.methods_by_name['UpdateWorkspaceDefinition']._serialized_options = b'\202\323\344\223\002U\032S/workspace_registry/api/v1/{id.org}/{id.project}/{id.domain}/{id.name}/{id.version}'
  _WORKSPACEREGISTRYSERVICE.methods_by_name['ListWorkspaceDefinitionEntities']._options = None
  _WORKSPACEREGISTRYSERVICE.methods_by_name['ListWorkspaceDefinitionEntities']._serialized_options = b'\220\002\001\202\323\344\223\002>\022</workspace_registry/api/v1/entities/{org}/{project}/{domain}'
  _WORKSPACEREGISTRYSERVICE.methods_by_name['ListWorkspaceDefinitions']._options = None
  _WORKSPACEREGISTRYSERVICE.methods_by_name['ListWorkspaceDefinitions']._serialized_options = b'\220\002\001\202\323\344\223\002s\0223/workspace_registry/api/v1/{org}/{project}/{domain}Z<\022:/workspace_registry/api/v1/{org}/{project}/{domain}/{name}'
  _globals['_WORKSPACEREGISTRYSERVICE']._serialized_start=145
  _globals['_WORKSPACEREGISTRYSERVICE']._serialized_end=1571
# @@protoc_insertion_point(module_scope)



import datetime
import typing
from typing import Tuple

import rich_click as click
from flytekit import BlobType, FlyteContext, Literal, StructuredDataset
from flytekit import LiteralType as lt
from flytekit.extend import TypeEngine
from flytekit.interaction import click_types
from flytekit.models.types import LiteralType, StructuredDatasetType
from flytekit.models.types import SimpleType as st
from flytekit.types.directory import FlyteDirectory
from flytekit.types.file import FlyteFile

from union.cli._common import CommandBase


class ArtifactCreateCommand(CommandBase):
    """
    This is a Command class for creating an artifact. This adds parameters for various artifact types,
    using flytekit's click types.

    TODO we have to add support for collections.
    """

    _ARTIFACT_FOR_TYPES = [
        {
            "param_name": "from_float",
            "description_extra": " (float)",
            "literal_type": lt(simple=st.FLOAT),
            "default_python_type": float,
        },
        {
            "param_name": "from_int",
            "description_extra": " (int)",
            "literal_type": lt(simple=st.INTEGER),
            "default_python_type": int,
        },
        {
            "param_name": "from_str",
            "description_extra": " (str)",
            "literal_type": lt(simple=st.STRING),
            "default_python_type": str,
        },
        {
            "param_name": "from_bool",
            "description_extra": " (bool)",
            "literal_type": lt(simple=st.BOOLEAN),
            "default_python_type": bool,
        },
        {
            "param_name": "from_datetime",
            "description_extra": " (datetime)",
            "literal_type": lt(simple=st.DATETIME),
            "default_python_type": datetime.datetime,
        },
        {
            "param_name": "from_duration",
            "description_extra": " (duration)",
            "literal_type": lt(simple=st.DURATION),
            "default_python_type": datetime.timedelta,
        },
        {
            "param_name": "from_json",
            "description_extra": " (struct)",
            "literal_type": lt(simple=st.STRUCT),
            "default_python_type": dict,
        },
        {
            "param_name": "from_dataframe",
            "description_extra": " (parquet)",
            "literal_type": lt(structured_dataset_type=StructuredDatasetType()),
            "default_python_type": StructuredDataset,
        },
        {
            "param_name": "from_file",
            "description_extra": " (file)",
            "literal_type": lt(blob=BlobType(dimensionality=BlobType.BlobDimensionality.SINGLE, format="")),
            "default_python_type": FlyteFile,
        },
        {
            "param_name": "from_dir",
            "description_extra": " (dir)",
            "literal_type": lt(blob=BlobType(dimensionality=BlobType.BlobDimensionality.MULTIPART, format="")),
            "default_python_type": FlyteDirectory,
        },
    ]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.params.extend(self._create_options())

    @classmethod
    def _create_options(cls) -> typing.List[click.Option]:
        options = []
        for v in cls._ARTIFACT_FOR_TYPES:
            ct = click_types.literal_type_to_click_type(v["literal_type"], v["default_python_type"])
            options.append(
                click.Option(
                    param_decls=[f"--{v['param_name']}"],
                    type=ct,
                    required=False,
                    help=f"Create an artifact of type {v['description_extra']}",
                )
            )
        return options

    @classmethod
    def get_literal_from_args(
        cls, flyte_ctx: FlyteContext, args: typing.Dict[str, typing.Any]
    ) -> Tuple[Literal, LiteralType]:
        for v in cls._ARTIFACT_FOR_TYPES:
            param_name = v["param_name"]
            if param_name in args:
                if args[param_name] is not None:
                    val = args[param_name]
                    lt = v["literal_type"]
                    # TODO this is a hack, need help to set the remote directory to None
                    # I think the click_types has a bug for Directory type
                    if isinstance(val, FlyteDirectory):
                        val._remote_directory = None
                    lit = TypeEngine.to_literal(flyte_ctx, val, v["default_python_type"], lt)
                    return lit.to_flyte_idl(), lt.to_flyte_idl()
        raise click.BadParameter("Artifact should be of one of the supported types, `--from_*`")



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: secret/payload.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from union.internal.secret import definition_pb2 as secret_dot_definition__pb2
from union.internal.validate.validate import validate_pb2 as validate_dot_validate__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x14secret/payload.proto\x12\x0f\x63loudidl.secret\x1a\x17secret/definition.proto\x1a\x17validate/validate.proto\"\x90\x01\n\x13\x43reateSecretRequest\x12;\n\x02id\x18\x01 \x01(\x0b\x32!.cloudidl.secret.SecretIdentifierB\x08\xfa\x42\x05\x8a\x01\x02\x10\x01R\x02id\x12<\n\x0bsecret_spec\x18\x02 \x01(\x0b\x32\x1b.cloudidl.secret.SecretSpecR\nsecretSpec\"\x16\n\x14\x43reateSecretResponse\"\x90\x01\n\x13UpdateSecretRequest\x12;\n\x02id\x18\x01 \x01(\x0b\x32!.cloudidl.secret.SecretIdentifierB\x08\xfa\x42\x05\x8a\x01\x02\x10\x01R\x02id\x12<\n\x0bsecret_spec\x18\x02 \x01(\x0b\x32\x1b.cloudidl.secret.SecretSpecR\nsecretSpec\"\x16\n\x14UpdateSecretResponse\"O\n\x10GetSecretRequest\x12;\n\x02id\x18\x01 \x01(\x0b\x32!.cloudidl.secret.SecretIdentifierB\x08\xfa\x42\x05\x8a\x01\x02\x10\x01R\x02id\"D\n\x11GetSecretResponse\x12/\n\x06secret\x18\x01 \x01(\x0b\x32\x17.cloudidl.secret.SecretR\x06secret\"R\n\x13\x44\x65leteSecretRequest\x12;\n\x02id\x18\x01 \x01(\x0b\x32!.cloudidl.secret.SecretIdentifierB\x08\xfa\x42\x05\x8a\x01\x02\x10\x01R\x02id\"\x16\n\x14\x44\x65leteSecretResponse\"\xc4\x02\n\x12ListSecretsRequest\x12\"\n\x0corganization\x18\x01 \x01(\tR\x0corganization\x12\x16\n\x06\x64omain\x18\x02 \x01(\tR\x06\x64omain\x12\x18\n\x07project\x18\x03 \x01(\tR\x07project\x12\x14\n\x05limit\x18\x04 \x01(\x05R\x05limit\x12\x14\n\x05token\x18\x05 \x01(\tR\x05token\x12g\n\x12per_cluster_tokens\x18\x06 \x03(\x0b\x32\x39.cloudidl.secret.ListSecretsRequest.PerClusterTokensEntryR\x10perClusterTokens\x1a\x43\n\x15PerClusterTokensEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01\"\x8d\x02\n\x13ListSecretsResponse\x12\x31\n\x07secrets\x18\x01 \x03(\x0b\x32\x17.cloudidl.secret.SecretR\x07secrets\x12\x14\n\x05token\x18\x02 \x01(\tR\x05token\x12h\n\x12per_cluster_tokens\x18\x03 \x03(\x0b\x32:.cloudidl.secret.ListSecretsResponse.PerClusterTokensEntryR\x10perClusterTokens\x1a\x43\n\x15PerClusterTokensEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01\x42\xad\x01\n\x13\x63om.cloudidl.secretB\x0cPayloadProtoH\x02P\x01Z)github.com/unionai/cloud/gen/pb-go/secret\xa2\x02\x03\x43SX\xaa\x02\x0f\x43loudidl.Secret\xca\x02\x0f\x43loudidl\\Secret\xe2\x02\x1b\x43loudidl\\Secret\\GPBMetadata\xea\x02\x10\x43loudidl::Secretb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'secret.payload_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\023com.cloudidl.secretB\014PayloadProtoH\002P\001Z)github.com/unionai/cloud/gen/pb-go/secret\242\002\003CSX\252\002\017Cloudidl.Secret\312\002\017Cloudidl\\Secret\342\002\033Cloudidl\\Secret\\GPBMetadata\352\002\020Cloudidl::Secret'
  _CREATESECRETREQUEST.fields_by_name['id']._options = None
  _CREATESECRETREQUEST.fields_by_name['id']._serialized_options = b'\372B\005\212\001\002\020\001'
  _UPDATESECRETREQUEST.fields_by_name['id']._options = None
  _UPDATESECRETREQUEST.fields_by_name['id']._serialized_options = b'\372B\005\212\001\002\020\001'
  _GETSECRETREQUEST.fields_by_name['id']._options = None
  _GETSECRETREQUEST.fields_by_name['id']._serialized_options = b'\372B\005\212\001\002\020\001'
  _DELETESECRETREQUEST.fields_by_name['id']._options = None
  _DELETESECRETREQUEST.fields_by_name['id']._serialized_options = b'\372B\005\212\001\002\020\001'
  _LISTSECRETSREQUEST_PERCLUSTERTOKENSENTRY._options = None
  _LISTSECRETSREQUEST_PERCLUSTERTOKENSENTRY._serialized_options = b'8\001'
  _LISTSECRETSRESPONSE_PERCLUSTERTOKENSENTRY._options = None
  _LISTSECRETSRESPONSE_PERCLUSTERTOKENSENTRY._serialized_options = b'8\001'
  _globals['_CREATESECRETREQUEST']._serialized_start=92
  _globals['_CREATESECRETREQUEST']._serialized_end=236
  _globals['_CREATESECRETRESPONSE']._serialized_start=238
  _globals['_CREATESECRETRESPONSE']._serialized_end=260
  _globals['_UPDATESECRETREQUEST']._serialized_start=263
  _globals['_UPDATESECRETREQUEST']._serialized_end=407
  _globals['_UPDATESECRETRESPONSE']._serialized_start=409
  _globals['_UPDATESECRETRESPONSE']._serialized_end=431
  _globals['_GETSECRETREQUEST']._serialized_start=433
  _globals['_GETSECRETREQUEST']._serialized_end=512
  _globals['_GETSECRETRESPONSE']._serialized_start=514
  _globals['_GETSECRETRESPONSE']._serialized_end=582
  _globals['_DELETESECRETREQUEST']._serialized_start=584
  _globals['_DELETESECRETREQUEST']._serialized_end=666
  _globals['_DELETESECRETRESPONSE']._serialized_start=668
  _globals['_DELETESECRETRESPONSE']._serialized_end=690
  _globals['_LISTSECRETSREQUEST']._serialized_start=693
  _globals['_LISTSECRETSREQUEST']._serialized_end=1017
  _globals['_LISTSECRETSREQUEST_PERCLUSTERTOKENSENTRY']._serialized_start=950
  _globals['_LISTSECRETSREQUEST_PERCLUSTERTOKENSENTRY']._serialized_end=1017
  _globals['_LISTSECRETSRESPONSE']._serialized_start=1020
  _globals['_LISTSECRETSRESPONSE']._serialized_end=1289
  _globals['_LISTSECRETSRESPONSE_PERCLUSTERTOKENSENTRY']._serialized_start=950
  _globals['_LISTSECRETSRESPONSE_PERCLUSTERTOKENSENTRY']._serialized_end=1017
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: workspace/workspace_instance_service.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.api import annotations_pb2 as google_dot_api_dot_annotations__pb2
from union.internal.workspace import workspace_instance_payload_pb2 as workspace_dot_workspace__instance__payload__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n*workspace/workspace_instance_service.proto\x12\x12\x63loudidl.workspace\x1a\x1cgoogle/api/annotations.proto\x1a*workspace/workspace_instance_payload.proto2\x99\x07\n\x18WorkspaceInstanceService\x12\xa6\x01\n\x16StartWorkspaceInstance\x12\x31.cloudidl.workspace.StartWorkspaceInstanceRequest\x1a\x32.cloudidl.workspace.StartWorkspaceInstanceResponse\"%\x82\xd3\xe4\x93\x02\x1f\"\x1a/workspace_instance/api/v1:\x01*\x12\xcc\x01\n\x14GetWorkspaceInstance\x12/.cloudidl.workspace.GetWorkspaceInstanceRequest\x1a\x30.cloudidl.workspace.GetWorkspaceInstanceResponse\"Q\x90\x02\x01\x82\xd3\xe4\x93\x02H\x12\x46/workspace_instance/api/v1/{id.org}/{id.project}/{id.domain}/{id.name}\x12\xa6\x01\n\x16ListWorkspaceInstances\x12\x31.cloudidl.workspace.ListWorkspaceInstancesRequest\x1a\x32.cloudidl.workspace.ListWorkspaceInstancesResponse\"%\x90\x02\x01\x82\xd3\xe4\x93\x02\x1c\x12\x1a/workspace_instance/api/v1\x12\xcf\x01\n\x15StopWorkspaceInstance\x12\x30.cloudidl.workspace.StopWorkspaceInstanceRequest\x1a\x31.cloudidl.workspace.StopWorkspaceInstanceResponse\"Q\x82\xd3\xe4\x93\x02K*F/workspace_instance/api/v1/{id.org}/{id.project}/{id.domain}/{id.name}:\x01*\x12\x89\x01\n\x17WatchWorkspaceInstances\x12\x32.cloudidl.workspace.WatchWorkspaceInstancesRequest\x1a\x33.cloudidl.workspace.WatchWorkspaceInstancesResponse\"\x03\x90\x02\x01\x30\x01\x42\xd0\x01\n\x16\x63om.cloudidl.workspaceB\x1dWorkspaceInstanceServiceProtoH\x02P\x01Z,github.com/unionai/cloud/gen/pb-go/workspace\xa2\x02\x03\x43WX\xaa\x02\x12\x43loudidl.Workspace\xca\x02\x12\x43loudidl\\Workspace\xe2\x02\x1e\x43loudidl\\Workspace\\GPBMetadata\xea\x02\x13\x43loudidl::Workspaceb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'workspace.workspace_instance_service_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\026com.cloudidl.workspaceB\035WorkspaceInstanceServiceProtoH\002P\001Z,github.com/unionai/cloud/gen/pb-go/workspace\242\002\003CWX\252\002\022Cloudidl.Workspace\312\002\022Cloudidl\\Workspace\342\002\036Cloudidl\\Workspace\\GPBMetadata\352\002\023Cloudidl::Workspace'
  _WORKSPACEINSTANCESERVICE.methods_by_name['StartWorkspaceInstance']._options = None
  _WORKSPACEINSTANCESERVICE.methods_by_name['StartWorkspaceInstance']._serialized_options = b'\202\323\344\223\002\037\"\032/workspace_instance/api/v1:\001*'
  _WORKSPACEINSTANCESERVICE.methods_by_name['GetWorkspaceInstance']._options = None
  _WORKSPACEINSTANCESERVICE.methods_by_name['GetWorkspaceInstance']._serialized_options = b'\220\002\001\202\323\344\223\002H\022F/workspace_instance/api/v1/{id.org}/{id.project}/{id.domain}/{id.name}'
  _WORKSPACEINSTANCESERVICE.methods_by_name['ListWorkspaceInstances']._options = None
  _WORKSPACEINSTANCESERVICE.methods_by_name['ListWorkspaceInstances']._serialized_options = b'\220\002\001\202\323\344\223\002\034\022\032/workspace_instance/api/v1'
  _WORKSPACEINSTANCESERVICE.methods_by_name['StopWorkspaceInstance']._options = None
  _WORKSPACEINSTANCESERVICE.methods_by_name['StopWorkspaceInstance']._serialized_options = b'\202\323\344\223\002K*F/workspace_instance/api/v1/{id.org}/{id.project}/{id.domain}/{id.name}:\001*'
  _WORKSPACEINSTANCESERVICE.methods_by_name['WatchWorkspaceInstances']._options = None
  _WORKSPACEINSTANCESERVICE.methods_by_name['WatchWorkspaceInstances']._serialized_options = b'\220\002\001'
  _globals['_WORKSPACEINSTANCESERVICE']._serialized_start=141
  _globals['_WORKSPACEINSTANCESERVICE']._serialized_end=1062
# @@protoc_insertion_point(module_scope)
